#!/usr/bin/env python3
"""
.. module cli.runner
   :platform: Unix, Windows, Mac, Linux
.. moduleauthor:: Chenxing Luo <chenxing.luo@columbia.edu>
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

import os
import pathlib
import time

from qha.calculator import Calculator, SamePhDOSCalculator, DifferentPhDOSCalculator
from qha.basic_io.out import save_x_tp, save_x_tv, save_to_output, make_starting_string, make_tp_info, make_ending_string
from qha.settings import from_yaml
from .handler import QHACommandHandler


class QHARunner(QHACommandHandler):
    def __init__(self):
        super().__init__()

        self.settings = None
        self.start_time = None

    def init_parser(self, parser):
        super().init_parser(parser)
        parser.add_argument('settings', type=str)

    def run(self, namespace):
        start_time_total = time.time()

        user_settings = {}
        file_settings = namespace.settings
        settings = from_yaml(file_settings)

        for key in ('input', 'calculation',
                    'thermodynamic_properties', 'static_only', 'energy_unit',
                    'T_MIN', 'NT', 'DT', 'DT_SAMPLE',
                    'P_MIN', 'NTV', 'DELTA_P', 'DELTA_P_SAMPLE',
                    'volume_ratio', 'order', 'p_min_modifier',
                    'T4FV', 'output_directory', 'high_verbosity'):
            try:
                user_settings.update({key: settings[key]})
            except KeyError:
                raise KeyError("Key '{0}' is not set in your settings!")

        if not os.path.exists(user_settings['output_directory']):
            os.makedirs(user_settings['output_directory'])

        user_settings.update({'qha_output': os.path.join(user_settings['output_directory'], 'output.txt')})

        try:
            os.remove(user_settings['qha_output'])
        except OSError:
            pass

        save_to_output(user_settings['qha_output'], make_starting_string())

        calculation_type = user_settings['calculation'].lower()
        if calculation_type == 'single':
            calc = Calculator(user_settings)
            print("You have single-configuration calculation assumed.")
        elif calculation_type == 'same phonon dos':
            calc = SamePhDOSCalculator(user_settings)
            print("You have multi-configuration calculation with the same phonon DOS assumed.")
        elif calculation_type == 'different phonon dos':
            calc = DifferentPhDOSCalculator(user_settings)
            print("You have multi-configuration calculation with different phonon DOS assumed.")
        else:
            raise ValueError("The 'calculation' in your settings in not recognized! It must be one of:"
                             "'single', 'same phonon dos', 'different phonon dos'!")

        save_to_output(user_settings['qha_output'],
                       make_tp_info(calc.temperature_array[0], calc.temperature_array[-1 - 4],
                                    calc.desired_pressures_gpa[0],
                                    calc.desired_pressures_gpa[-1]))

        calc.read_input()

        print("Caution: If negative frequencies found, they are currently treated as 0!")
        tmp = calc.where_negative_frequencies
        if tmp is not None and not (tmp.T[-1].max() <= 2):  # Don't delete this parenthesis!
            if calc.frequencies.ndim == 4:  # Multiple configuration
                for indices in tmp:
                    print(
                        "Found negative frequency in {0}th configuration {1}th volume {2}th q-point {3}th band".format(
                            *tuple(indices + 1)))
            elif calc.frequencies.ndim == 3:  # Single configuration
                for indices in tmp:
                    print(
                        "Found negative frequency in {0}th volume {1}th q-point {2}th band".format(*tuple(indices + 1)))

        calc.refine_grid()

        if user_settings['high_verbosity']:
            save_to_output(user_settings['qha_output'],
                           'The volume range used in this calculation expanded x {0:6.4f}'.format(calc.v_ratio))

        calc.desired_pressure_status()

        temperature_array = calc.temperature_array
        desired_pressures_gpa = calc.desired_pressures_gpa
        temperature_sample = calc.temperature_sample_array
        p_sample_gpa = calc.pressure_sample_array

        results_folder = pathlib.Path(user_settings['output_directory'])

        calculation_option = {'F': 'f_tp',
                              'G': 'g_tp',
                              'H': 'h_tp',
                              'U': 'u_tp',
                              'V': 'v_tp',
                              'Cv': 'cv_tp_jmolk',
                              'Cp': 'cp_tp_jmolk',
                              'Bt': 'bt_tp_gpa',
                              'Btp': 'btp_tp',
                              'Bs': 'bs_tp_gpa',
                              'alpha': 'alpha_tp',
                              'gamma': 'gamma_tp',
                              }

        file_ftv_fitted = results_folder / 'f_tv_fitted_ev_ang3.txt'
        save_x_tv(calc.f_tv_ev, temperature_array, calc.finer_volumes_ang3, temperature_sample, file_ftv_fitted)

        file_ftv_non_fitted = results_folder / 'f_tv_nonfitted_ev_ang3.txt'
        save_x_tv(calc.vib_ev, temperature_array, calc.volumes_ang3, temperature_sample, file_ftv_non_fitted)

        file_ptv_gpa = results_folder / 'p_tv_gpa.txt'
        save_x_tv(calc.p_tv_gpa, temperature_array, calc.finer_volumes_ang3, temperature_sample, file_ptv_gpa)

        for idx in calc.settings['thermodynamic_properties']:
            if idx in ['F', 'G', 'H', 'U']:
                attr_name = calculation_option[idx] + '_' + calc.settings['energy_unit']
                file_name = attr_name + '.txt'
                file_dir = results_folder / file_name
                save_x_tp(getattr(calc, attr_name), temperature_array, desired_pressures_gpa, p_sample_gpa, file_dir)

            if idx == 'V':
                v_bohr3 = calculation_option[idx] + '_' + 'bohr3'
                file_name_bohr3 = v_bohr3 + '.txt'
                file_dir_au = results_folder / file_name_bohr3
                v_ang3 = calculation_option[idx] + '_' + 'ang3'
                file_name_ang3 = v_ang3 + '.txt'
                file_dir_ang3 = results_folder / file_name_ang3

                save_x_tp(getattr(calc, v_bohr3), temperature_array, desired_pressures_gpa, p_sample_gpa, file_dir_au)
                save_x_tp(getattr(calc, v_ang3), temperature_array, desired_pressures_gpa, p_sample_gpa, file_dir_ang3)

            if idx in ['Cv', 'Cp', 'Bt', 'Btp', 'Bs', 'alpha', 'gamma']:
                attr_name = calculation_option[idx]
                file_name = attr_name + '.txt'
                file_dir = results_folder / file_name
                save_x_tp(getattr(calc, attr_name), temperature_array, desired_pressures_gpa, p_sample_gpa, file_dir)

        end_time_total = time.time()
        time_elapsed = end_time_total - start_time_total
        save_to_output(user_settings['qha_output'], make_ending_string(time_elapsed))



#!/usr/bin/env python3
"""
.. module output
   :platform: Unix, Windows, Mac, Linux
   :synopsis: doc
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

import textwrap
from datetime import datetime

import pandas as pd


def save_to_output(fn_output, text):
    with open(fn_output, 'a') as f:
        f.write(text + '\n')


def save_x_tp(df, t, desired_pressures_gpa, p_sample_gpa, outfile_name):
    # To fix the last 2 temperature points of Cp is not accurate, we added 4 temperature points,
    # Now to get rid of the added temperature points before saving calculated properties into files.
    df = pd.DataFrame(df, index=t, columns=desired_pressures_gpa)[:-4]
    df.columns.name = 'T(K)\P(GPa)'
    sample = df[p_sample_gpa]
    with open(outfile_name, 'w') as f:
        f.write(sample.to_string())


def save_x_pt(df, t, desired_pressures_gpa, t_sample, outfile_name):
    # To fix the last 2 temperature points of Cp is not accurate, we added 4 temperature points,
    # Now to get rid of the added temperature points before saving calculated properties into files.
    df = pd.DataFrame(df[:-4].T, index=desired_pressures_gpa, columns=t[:-4])
    df.columns.name = 'P(GPa)\T(K)'
    sample = df[t_sample]
    with open(outfile_name, 'w') as f:
        f.write(sample.to_string())


def save_x_vt(x, t, volume_grid, t_sample, outfile_name):
    df = pd.DataFrame(x.T, index=volume_grid, columns=t)
    df.columns.name = 'V(A^3)\T(K)'
    sample = df[t_sample]
    with open(outfile_name, 'w') as f:
        f.write(sample.to_string())


def save_x_tv(x, t, volume_grid, t_sample, outfile_name):
    df = pd.DataFrame(x, index=t, columns=volume_grid)[:-4]
    df.columns.name = 'T(K)\V(A^3)'
    sample = df.T[t_sample[:-4]].T
    with open(outfile_name, 'w') as f:
        f.write(sample.to_string())


def make_starting_string() -> str:
    return textwrap.dedent("""\
        ============================================================
        Current time: {0:%Y-%m-%d %H:%M:%S}
        """.format(datetime.utcnow()))


def make_tp_info(min_temperature, max_temperature, min_pressure, max_pressure):
    return textwrap.dedent("""\
        ------------------------------------------------------------
         Desired T range:    {0:6.2f} to {1:6.2f}  K
         Desired P range:    {2:6.2f} to {3:6.2f}  GPa
        ------------------------------------------------------------
        """.format(min_temperature, max_temperature, min_pressure, max_pressure))


def make_ending_string(time_elapsed) -> str:
    return textwrap.dedent("""\
        ------------------------------------------------------------
        Total elapsed time is: {0:8.2f} seconds
        Thanks for using QHA code, have a nice one :)
        ============================================================
        """.format(time_elapsed))



#!/usr/bin/env python3
"""
.. module single_configuration
   :platform: Unix, Windows, Mac, Linux
   :synopsis: Free energy calculator for a single configuration.
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

import numpy as np
from lazy_property import LazyProperty
from numba import jit, float64, boolean

from qha.statmech import ho_free_energy
from qha.type_aliases import Scalar, Vector, Matrix, Array3D

# ===================== What can be exported? =====================
__all__ = ['free_energy', 'HOFreeEnergySampler']


@jit(float64[:](float64, float64[:], float64[:], float64[:, :, :], boolean), cache=True)
def free_energy(temperature: Scalar, q_weights: Vector, static_energies: Vector, frequencies: Array3D,
                static_only: bool = False) -> Vector:
    """
    The total free energy at a certain temperature.

    :param temperature: A ``float`` that represents the temperature at which the total free energy is calculated. This
        value is in unit 'Kelvin'.
    :param q_weights: An :math:`m \\times 1` vector that represents the weight of each q-point in Brillouin zone
        sampling. This vector can be non-normalized since a normalization will be done internally.
    :param static_energies: An :math:`n \\times 1` vector that represents the static energy of the system with
        :math:`n` different volumes. This should be the same unit as user-defined in the "settings" file.
    :param frequencies: An :math:`n \\times m \\times l` 3D array that represents the frequency read from file.
    :param static_only: If ``True``, directly return the static energies, i.e., the *static_energies* parameter itself.
        This is useful when the user just wants to see static result while keeping all other functions unchanged.
    :return: An :math:`n \\times 1` vector that represents the total free energy of the system with :math:`n` different
        volumes. The default unit is the same as in function ``ho_free_energy``.
    """
    if not np.all(np.greater_equal(q_weights, 0)):
        raise ValueError('Weights should all be greater equal than 0!')

    if static_only:
        return static_energies

    scaled_q_weights: Vector = q_weights / np.sum(q_weights)
    vibrational_energies: Vector = np.dot(ho_free_energy(temperature, frequencies).sum(axis=2), scaled_q_weights)
    return static_energies + vibrational_energies


class HOFreeEnergySampler:
    """
    A harmonic oscillator free energy sampler. The differences between ``free_energy`` and ``HOFreeEnergySampler`` are

    - ``HOFreeEnergySampler`` provides several ways to compute free energies on different q-points, bands or
        volumes, which could be used to check the correctness of the data.
    - ``HOFreeEnergySampler`` can make calculation lazy, rather than directly calculates the values immediately.
    - ``free_energy`` is usually enough for computing the final free energy results. ``HOFreeEnergySampler``
      only calculates its vibrational part, static energies are not taken into account.

    :param temperature: A ``float`` that represents the temperature at which the total free energy is calculated. This
        value is in unit 'Kelvin'.
    :param q_weights: An :math:`m \\times 1` vector that represents the weight of each q-point in Brillouin zone
        sampling. This vector can be non-normalized since a normalization will be done internally.
    :param frequencies: An :math:`n \\times m \\times l` 3D array that represents the frequencies read from a file,
        usually in unit :math:`\\text{cm}^{-1}`.
    """

    def __init__(self, temperature: float, q_weights: Vector, frequencies: Array3D):
        self.temperature = temperature  # Fix temperature and volume, just sample q-points and bands
        self.q_weights = np.array(q_weights, dtype=float)
        self.omegas = np.array(frequencies, dtype=float)
        if not np.all(np.greater_equal(q_weights, 0)):
            raise ValueError('Weights should all be greater equal than 0!')
        else:
            self._scaled_q_weights = q_weights / np.sum(q_weights)

    def on_q_point(self, i: int) -> Matrix:  # E1(i,m)
        """
        Sample free energy on the :math:`i` th q-point.

        :param i: An integer labeling :math:`i` th q-point.
        :return: The accumulated free energy on the :math:`i` th q-point.
        """
        return ho_free_energy(self.temperature, self.omegas[:, i, :])

    def on_band(self, i: int) -> Matrix:
        """
        Sample free energy on the :math:`i` th band.

        :param i: An integer labeling :math:`i` th band.
        :return: The accumulated free energy on the :math:`i`th q-point.
        """
        return ho_free_energy(self.temperature, self.omegas[:, :, i])

    def on_volume(self, i: int) -> Scalar:
        """
        Sample free energy on the :math:`i` th volume.

        :param i: An integer labeling :math:`i` th volume.
        :return: The accumulated free energy on the :math:`i` th volume.
        """
        return np.vdot(ho_free_energy(self.temperature, self.omegas[i]).sum(axis=1), self._scaled_q_weights)

    @LazyProperty
    def on_all_volumes(self) -> Vector:
        """
        Sample free energies on every volume.

        :return: A vector with length equals the number of volumes. Each element is the free energy of
            one volume.
        """
        # First calculate free energies on a 3D array, then sum along the third axis (bands),
        # at last contract weights and free energies on all q-points.
        return np.dot(ho_free_energy(self.temperature, self.omegas).sum(axis=2), self._scaled_q_weights)



#!/usr/bin/env python3
"""
.. module cli.handler
   :platform: Unix, Windows, Mac, Linux
.. moduleauthor:: Chenxing Luo <chenxing.luo@columbia.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

import abc
import argparse


class QHACommandHandler(abc.ABC):
    @abc.abstractmethod
    def init_parser(self, parser: argparse.ArgumentParser):
        ...

    @abc.abstractmethod
    def run(self, namespace):
        ...



#!/usr/bin/env python3
"""
.. module cli.plotter
   :platform: Unix, Windows, Mac, Linux
.. moduleauthor:: Chenxing Luo <chenxing.luo@columbia.edu>
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

import os
import pathlib

import qha.tools
from qha.cli.handler import QHACommandHandler
from qha.plotting import Plotter
from qha.settings import from_yaml


class QHAPlotter(QHACommandHandler):
    def __init__(self):
        super().__init__()

    def init_parser(self, parser):
        super().init_parser(parser)
        parser.add_argument('settings', type=str)
        parser.add_argument('--outdir', type=str, help='output directory')

    def run(self, namespace):
        user_settings = {}  # save necessary info for plotting later
        file_settings = namespace.settings
        settings = from_yaml(file_settings)

        for key in ('energy_unit', 'NT', 'DT', 'DT_SAMPLE', 'P_MIN', 'NTV', 'DELTA_P', 'DELTA_P_SAMPLE',
                    'thermodynamic_properties', 'T4FV', 'output_directory'):
            try:
                user_settings.update({key: settings[key]})
            except KeyError:
                raise KeyError("Key '{0}' is not set in your settings!")

        if not os.path.exists(user_settings['output_directory']):
            raise FileNotFoundError("There is no results folder, please run: `qha-run` first! ")

        plotter = Plotter(user_settings)

        desired_pressures_gpa = qha.tools.arange(user_settings['P_MIN'], user_settings['NTV'], user_settings['DELTA_P'])
        user_settings.update({'DESIRED_PRESSURES_GPa': desired_pressures_gpa})

        results_folder = pathlib.Path(user_settings['output_directory'])

        calculation_option = {'F': 'f_tp',
                              'G': 'g_tp',
                              'H': 'h_tp',
                              'U': 'u_tp',
                              'V': 'v_tp',
                              'Cv': 'cv_tp_jmolk',
                              'Cp': 'cp_tp_jmolk',
                              'Bt': 'bt_tp_gpa',
                              'Btp': 'btp_tp',
                              'Bs': 'bs_tp_gpa',
                              'alpha': 'alpha_tp',
                              'gamma': 'gamma_tp',
                              }

        file_ftv_fitted = results_folder / 'f_tv_fitted_ev_ang3.txt'
        user_settings.update({'f_tv_fitted': file_ftv_fitted})

        file_ftv_non_fitted = results_folder / 'f_tv_nonfitted_ev_ang3.txt'
        user_settings.update({'f_tv_non_fitted': file_ftv_non_fitted})

        file_ptv_gpa = results_folder / 'p_tv_gpa.txt'
        user_settings.update({'p_tv_gpa': file_ptv_gpa})

        plotter.fv_pv()

        for idx in user_settings['thermodynamic_properties']:
            if idx in ['F', 'G', 'H', 'U']:
                attr_name = calculation_option[idx] + '_' + user_settings['energy_unit']
                file_name = attr_name + '.txt'
                file_dir = results_folder / file_name
                user_settings.update({idx: file_dir})
                plotter.plot2file(idx)

            if idx == 'V':
                v_ang3 = calculation_option[idx] + '_' + 'ang3'
                file_name_ang3 = v_ang3 + '.txt'
                file_dir_ang3 = results_folder / file_name_ang3
                user_settings.update({idx: file_dir_ang3})
                plotter.plot2file(idx)

            if idx in ['Cv', 'Cp', 'Bt', 'Btp', 'Bs', 'alpha', 'gamma']:
                attr_name = calculation_option[idx]
                file_name = attr_name + '.txt'
                file_dir = results_folder / file_name
                user_settings.update({idx: file_dir})
                plotter.plot2file(idx)



#!/usr/bin/env python3
"""
.. module multi_configurations.different_phonon_dos
   :platform: Unix, Windows, Mac, Linux
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
"""

from typing import Optional

import numpy as np
from lazy_property import LazyProperty
from scipy.constants import physical_constants as pc
from scipy.special import logsumexp

import qha.settings
from qha.single_configuration import free_energy
from qha.tools import calibrate_energy_on_reference
from qha.type_aliases import Array4D, Scalar, Vector, Matrix

# ===================== What can be exported? =====================
__all__ = ['PartitionFunction']

K = {'ha': pc['Boltzmann constant in eV/K'][0] / pc['Hartree energy in eV'][0],
     'ry': pc['Boltzmann constant in eV/K'][0] / pc['Rydberg constant times hc in eV'][0],
     'ev': pc['Boltzmann constant in eV/K'][0],
     'SI': pc['Boltzmann constant'][0]}[qha.settings.energy_unit]


class PartitionFunction:
    """
    A class that represents the partition function of multiple configurations with different phonon density of states.
    In mathematics, it is represented as

    .. math::

       Z_{\\text{all configs}}(T, V) = \sum_{j = 1}^{N_{c}} g_{j} Z_{j}(T, V),

    where :math:`N_{c}` stands for the number of configurations and :math:`g_{j}` stands for degeneracy for the
    :math:`j` th configuration.

    :param temperature: The temperature at which the partition function is calculated.
    :param degeneracies: An array of degeneracies of each configuration, which will not be normalized in the
        calculation. They should all be positive integers.
    :param q_weights: The weights of all q-points that are sampled, can be a 2D matrix so each configuration can have
        a little bit different q-point weights, but the number of q-points of each configuration must be the same.
    :param static_energies: The static energy of each configuration of each volume.
    :param volumes: A matrix of volumes of each configuration, should have the same values for each
        configuration.
    :param frequencies: A 4D array that specifies the frequency on each configuration, volume, q-point and mode.
    :param static_only: Whether the calculation only takes static contribution and does not consider
        the vibrational contribution, by default, is ``False``.
    :param precision: The precision of a big float number to represent the partition function since it is a very large
        value, by default, is ``500``.
    :param order: The order of Birch--Murnaghan equation of state fitting, by default, is ``3``.
    """

    def __init__(self, temperature: Scalar, degeneracies: Vector, q_weights: Matrix, static_energies: Matrix,
                 volumes: Matrix, frequencies: Array4D, static_only: Optional[bool] = False,
                 precision: Optional[int] = 500, order: Optional[int] = 3):
        if not np.all(np.greater_equal(degeneracies, 0)):
            raise ValueError('Degeneracies should all be greater equal than 0!')
        if not np.all(np.greater_equal(
                q_weights, 0)):  # Weights should all be greater equal than 0, otherwise sum will be wrong.
            raise ValueError('Weights should all be greater equal than 0!')

        self.frequencies = np.array(frequencies)
        if self.frequencies.ndim != 4:
            raise ValueError("*frequencies* must be a four-dimensional array!")

        self.static_energies = np.array(static_energies)
        if self.static_energies.ndim != 2:
            raise ValueError("*static_energies* must be a two-dimensional array!")

        if temperature < 1e-1:
            self.temperature = 1
        else:
            self.temperature = temperature
        self.degeneracies = np.array(degeneracies)
        self.q_weights = np.array(q_weights)
        self._scaled_q_weights = self.q_weights / np.sum(q_weights)
        self.volumes = volumes
        self.static_only = static_only
        self.precision = int(precision)
        self.order = int(order)

    @LazyProperty
    def unaligned_free_energies_for_each_configuration(self):
        """
        If the input free energy is not aligned for each configuration, first just calculate the "raw"
        free energy on each volume and each configuration.

        :return: A matrix, the "raw" free energy of each configuration of each volume.
        """
        configurations_amount, _ = self.volumes.shape
        return np.array([free_energy(self.temperature, self.q_weights[i], self.static_energies[i], self.frequencies[i],
                                     self.static_only) for i in range(configurations_amount)])

    @LazyProperty
    def aligned_free_energies_for_each_configuration(self):
        """
        Then do a fitting to align all these free energies.

        :return: A matrix, the aligned free energy of each configuration of each volume.
        """
        return calibrate_energy_on_reference(self.volumes, self.unaligned_free_energies_for_each_configuration,
                                             self.order)

    @LazyProperty
    def partition_functions_for_each_configuration(self):
        """
        Inversely solve the free energy to get partition function for :math:`j` th configuration by

        .. math::

           Z_{j}(T, V) = \exp \\bigg( -\\frac{ F_{j}(T, V) }{ k_B T } \\bigg).

        :return: A matrix, the partition function of each configuration of each volume.
        """
        try:
            import bigfloat
        except ImportError:
            raise ImportError("Install ``bigfloat`` package to use {0} object!".format(self.__class__.__name__))

        with bigfloat.precision(self.precision):
            return np.array([bigfloat.exp(d) for d in  # shape = (# of volumes for each configuration, 1)
                             logsumexp(-self.aligned_free_energies_for_each_configuration.T / (K * self.temperature),
                                       axis=1, b=self.degeneracies)])

    def get_free_energies(self):
        """
        The free energy calculated from the partition function :math:`Z_{\\text{all configs}}(T, V)` by

        .. math::

           F_{\\text{all configs}}(T, V) = - k_B T \ln Z_{\\text{all configs}}(T, V).

        :return: The free energy on a temperature-volume grid.
        """
        try:
            import bigfloat
        except ImportError:
            raise ImportError("Install ``bigfloat`` package to use {0} object!".format(self.__class__.__name__))

        with bigfloat.precision(self.precision):
            log_z = np.array([bigfloat.log(d) for d in self.partition_functions_for_each_configuration], dtype=float)
        return -K * self.temperature * log_z



#!/usr/bin/env python3
"""
.. module settings
   :platform: Unix, Windows, Mac, Linux
   :synopsis: In this module, we give a simple computational "settings" ``DEFAULT_SETTINGS``, which will
    be used by default in the following calculation, if the users do not specify "settings" themselves.
    If they want, they can write some settings in a YAML file.
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
"""

import collections
from typing import Union, Tuple, Dict, Any

import yaml

# ===================== What can be exported? =====================
__all__ = ['DEFAULT_SETTINGS', 'Settings', 'from_yaml']

DEFAULT_SETTINGS: Dict[str, Any] = {
    'energy_unit': 'ry',
    'length_unit': 'angstrom',
    'order': 3,  # BM fitting order, can be 3, 4 or 5, normally, 3rd order is sufficient.
    'p_min_modifier': 1.0,
    'target': 'parallel',
    'T_MIN': 0,
    'DT_SAMPLE': 10,
    'DELTA_P': 0.1,
    'DELTA_P_SAMPLE': 1,
    'static_only': False,
    'volume_ratio': 1.45,
    # output setting
    'output_directory': './results/',
    'T4FV': ['0', '300'],
    'high_verbosity': False
}


class Settings(collections.ChainMap):
    """
    Contains both user "settings" and default "settings", where user "settings" will override default "settings".

    The default "settings" are

    .. code-block:: python

        DEFAULT_SETTINGS: = {
            'energy_unit': 'ry',
            'length_unit': 'angstrom',
            'order': 3,
            'p_min_modifier': 1.0,
            'target': 'parallel',
            'T_MIN': 0,
            'DT_SAMPLE': 10,
            'DELTA_P': 0.1,
            'DELTA_P_SAMPLE': 1,
            'static_only': False,
            'volume_ratio': 1.45,
            'output_directory': './results/',
            'T4FV': ['0', '300'],
            'high_verbosity': False
        }

    :param user_settings: It can be a single dictionary or a tuple of dictionaries, with strings as their keys.
    """

    def __init__(self, *user_settings: Union[Dict[str, Any], Tuple[Dict[str, Any], ...]]):
        super().__init__(*user_settings, DEFAULT_SETTINGS)

    def to_yaml_file(self, filename: str):
        if not filename.endswith('.yaml'):
            filename += '.yaml'
        with open(filename, 'w') as f:
            yaml.dump(self.maps, f)


def from_yaml(filename: str) -> Settings:
    """
    This function reads user "settings" from a YAML file, and generate a ``Settings`` class for later use.

    :param filename: The name of the YAML file.
    :return: A ``Settings`` class.
    """
    with open(filename, 'r') as f:
        return Settings(yaml.load(f))


global energy_unit
energy_unit = DEFAULT_SETTINGS['energy_unit']



#!/usr/bin/env python3
"""
.. module calculator
   :platform: Unix, Windows, Mac, Linux
   :synopsis: This is one of the most important modules in this package. It defines 3 classes, ``Calculator``,
    ``SamePhDOSCalculator``, and ``DifferentPhDOSCalculator``. The first one can be used to do single configuration
    calculation, the others are for multiple configurations calculation. Among them, the second one is for
    assuming all the configurations have the same phonon density of states, and the third one is for assuming
    all the configurations have different phonon density of states.
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

import textwrap
from typing import Dict, Any, Optional

import numpy as np
from lazy_property import LazyProperty

import qha.multi_configurations.different_phonon_dos as different_phonon_dos
import qha.multi_configurations.same_phonon_dos as same_phonon_dos
import qha.tools
from qha.grid_interpolation import FinerGrid
from qha.basic_io.out import save_to_output
from qha.basic_io import read_input
from qha.single_configuration import free_energy
from qha.thermodynamics import *
from qha.type_aliases import Vector
from qha.unit_conversion import gpa_to_ry_b3, ry_b3_to_gpa, b3_to_a3, ry_to_j_mol, ry_to_ev
from qha.v2p import v2p

# ===================== What can be exported? =====================
__all__ = ['Calculator', 'SamePhDOSCalculator', 'DifferentPhDOSCalculator']


class Calculator:
    def __init__(self, user_settings: Dict[str, Any]):
        runtime_settings = dict()

        allowed_keys = ('input', 'calculation',
                        'thermodynamic_properties', 'static_only', 'energy_unit',
                        'T_MIN', 'NT', 'DT', 'DT_SAMPLE',
                        'P_MIN', 'NTV', 'DELTA_P', 'DELTA_P_SAMPLE',
                        'volume_ratio', 'order', 'p_min_modifier',
                        'T4FV', 'output_directory', 'high_verbosity', 'qha_output')

        for key in allowed_keys:
            try:
                runtime_settings.update({key: user_settings[key]})
            except KeyError:
                continue  # If a key is not set in user settings, use the default.

        self._settings = runtime_settings

        self._formula_unit_number = None
        self._volumes = None
        self._static_energies = None
        self._frequencies = None
        self._q_weights = None

        self._finer_volumes_bohr3 = None
        self._f_tv_ry = None
        self._v_ratio = None

    @property
    def settings(self) -> Dict[str, Any]:
        """
        This is a read-only property.

        :return: The computational settings given by the user, combined with default settings. The user settings
            will cover the default settings.
        """
        return self._settings

    @property
    def formula_unit_number(self) -> Optional[int]:
        return self._formula_unit_number

    @property
    def volumes(self):
        return self._volumes

    @property
    def static_energies(self):
        return self._static_energies

    @property
    def frequencies(self):
        return self._frequencies

    @property
    def q_weights(self):
        return self._q_weights

    @property
    def finer_volumes_bohr3(self):
        return self._finer_volumes_bohr3

    @property
    def f_tv_ry(self):
        return self._f_tv_ry

    @property
    def v_ratio(self) -> Optional[float]:
        return self._v_ratio

    def read_input(self):
        try:
            formula_unit_number, volumes, static_energies, frequencies, q_weights = read_input(self.settings['input'])
        except KeyError:
            raise KeyError("The 'input' option must be given in your settings!")

        if not qha.tools.is_monotonic_decreasing(volumes):
            raise RuntimeError("Check the input file to make sure the volume decreases!")

        self._formula_unit_number: int = formula_unit_number
        self._volumes = volumes
        self._static_energies = static_energies
        self._frequencies = frequencies
        self._q_weights = q_weights

    @LazyProperty
    def where_negative_frequencies(self) -> Optional[Vector]:
        """
        The indices of negative frequencies are indicated.

        :return:
        """
        if self._frequencies is None:
            print("Please invoke ``read_input`` method first!")  # ``None`` is returned
        else:
            _ = np.transpose(np.where(self._frequencies < 0))
            if _.size == 0:
                return None

            return _

    @LazyProperty
    def temperature_array(self) -> Vector:
        """
        The temperature calculated to derive free energy at that temperature.

        :return:
        """
        # Normally, the last 2 temperature points in Cp are not accurate.
        # Here 4 more points are added for calculation, but they will be removed at the output files.
        minimum_temperature = self.settings['T_MIN']
        if minimum_temperature < 0:
            raise ValueError("Minimum temperature should be no less than 0!")

        return qha.tools.arange(minimum_temperature, self.settings['NT'] + 4, self.settings['DT'])

    def refine_grid(self):
        d = self.settings

        try:
            p_min, p_min_modifier, ntv, order = d['P_MIN'], d['p_min_modifier'], d['NTV'], d['order']
        except KeyError:
            raise KeyError("All the 'P_MIN', 'p_min_modifier', 'NTV', 'order' options must be given in your settings!")

        r = FinerGrid(p_min - p_min_modifier, ntv, order=order)

        if 'volume_ratio' in d:
            self._finer_volumes_bohr3, self._f_tv_ry, self._v_ratio = r.refine_grid(self.volumes, self.vib_ry,
                                                                                    ratio=d['volume_ratio'])
        else:
            self._finer_volumes_bohr3, self._f_tv_ry, self._v_ratio = r.refine_grid(self.volumes, self.vib_ry)

    @LazyProperty
    def vib_ry(self):
        # We grep all the arguments once since they are being invoked for thousands of times, and will be an overhead.
        args = self.q_weights, self.static_energies, self.frequencies, self.settings['static_only']

        mat = np.empty((self.temperature_array.size, self.volumes.size))
        for i, t in enumerate(self.temperature_array):
            mat[i] = free_energy(t, *(arg for arg in args))

        return mat

    @LazyProperty
    def thermodynamic_potentials(self) -> Dict[str, Any]:
        return thermodynamic_potentials(self.temperature_array, self.finer_volumes_bohr3, self.f_tv_ry, self.p_tv_au)

    @LazyProperty
    def temperature_sample_array(self):
        return self.temperature_array[0::int(self.settings['DT_SAMPLE'] / self.settings['DT'])]

    @LazyProperty
    def desired_pressures_gpa(self):
        d = self.settings
        return qha.tools.arange(d['P_MIN'], d['NTV'], d['DELTA_P'])

    @LazyProperty
    def desired_pressures(self):
        return gpa_to_ry_b3(self.desired_pressures_gpa)

    @LazyProperty
    def pressure_sample_array(self):
        return self.desired_pressures_gpa[0::int(self.settings['DELTA_P_SAMPLE'] / self.settings['DELTA_P'])]

    @LazyProperty
    def finer_volumes_ang3(self):
        return b3_to_a3(self.finer_volumes_bohr3)

    @LazyProperty
    def vib_ev(self):
        return ry_to_ev(self.vib_ry)

    @LazyProperty
    def volumes_ang3(self):
        return b3_to_a3(self.volumes)

    def desired_pressure_status(self) -> None:
        d = self.settings

        if d['high_verbosity']:
            save_to_output(d['qha_output'], "The pressure range can be dealt with: [{0:6.2f} to {1:6.2f}] GPa".format(
                self.p_tv_gpa[:, 0].max(), self.p_tv_gpa[:, -1].min()))

        if self.p_tv_gpa[:, -1].min() < self.desired_pressures_gpa.max():
            ntv_max = int((self.p_tv_gpa[:, -1].min() - self.desired_pressures_gpa.min()) / d['DELTA_P'])

            save_to_output(d['qha_output'], textwrap.dedent("""\
                           !!!ATTENTION!!!

                           DESIRED PRESSURE is too high (NTV is too large)!
                           QHA results might not be right!
                           Please reduce the NTV accordingly, for example, try to set NTV < {:4d}.
                           """.format(ntv_max)))

            raise ValueError("DESIRED PRESSURE is too high (NTV is too large), qha results might not be right!")

    @LazyProperty
    def p_tv_au(self):
        return pressure(self.finer_volumes_bohr3, self.f_tv_ry)

    @LazyProperty
    def f_tv_ev(self):
        return ry_to_ev(self.f_tv_ry)

    @LazyProperty
    def p_tv_gpa(self):
        return ry_b3_to_gpa(self.p_tv_au)

    @LazyProperty
    def f_tp_ry(self):
        return v2p(self.f_tv_ry, self.p_tv_au, self.desired_pressures)

    @LazyProperty
    def f_tp_ev(self):
        return ry_to_ev(self.f_tp_ry)

    @LazyProperty
    def u_tv_ry(self):
        return self.thermodynamic_potentials['U']

    @LazyProperty
    def u_tp_ry(self):
        return v2p(self.u_tv_ry, self.p_tv_au, self.desired_pressures)

    @LazyProperty
    def u_tp_ev(self):
        return ry_to_ev(self.u_tp_ry)

    @LazyProperty
    def h_tv_ry(self):
        return self.thermodynamic_potentials['H']

    @LazyProperty
    def h_tp_ry(self):
        return v2p(self.h_tv_ry, self.p_tv_au, self.desired_pressures)

    @LazyProperty
    def h_tp_ev(self):
        return ry_to_ev(self.h_tp_ry)

    @LazyProperty
    def g_tv_ry(self):
        return self.thermodynamic_potentials['G']

    @LazyProperty
    def g_tp_ry(self):
        return v2p(self.g_tv_ry, self.p_tv_au, self.desired_pressures)

    @LazyProperty
    def g_tp_ev(self):
        return ry_to_ev(self.g_tp_ry)

    @LazyProperty
    def bt_tv_au(self):
        return isothermal_bulk_modulus(self.finer_volumes_bohr3, self.p_tv_au)

    @LazyProperty
    def bt_tp_au(self):
        return v2p(self.bt_tv_au, self.p_tv_au, self.desired_pressures)

    @LazyProperty
    def bt_tp_gpa(self):
        return ry_b3_to_gpa(self.bt_tp_au)

    @LazyProperty
    def btp_tp(self):
        return bulk_modulus_derivative(self.desired_pressures, self.bt_tp_au)

    @LazyProperty
    def v_tp_bohr3(self):
        return volume(self.finer_volumes_bohr3, self.desired_pressures, self.p_tv_au)

    @LazyProperty
    def v_tp_ang3(self):
        return b3_to_a3(self.v_tp_bohr3)

    @LazyProperty
    def alpha_tp(self):
        return thermal_expansion_coefficient(self.temperature_array, self.v_tp_bohr3)

    @LazyProperty
    def cv_tv_au(self):
        return volumetric_heat_capacity(self.temperature_array, self.u_tv_ry)

    @LazyProperty
    def cv_tp_au(self):
        return v2p(self.cv_tv_au, self.p_tv_au, self.desired_pressures)

    @LazyProperty
    def cv_tp_jmolk(self):
        return ry_to_j_mol(self.cv_tp_au) / self.formula_unit_number

    @LazyProperty
    def gamma_tp(self):
        return gruneisen_parameter(self.v_tp_bohr3, self.bt_tp_au, self.alpha_tp, self.cv_tp_au)

    @LazyProperty
    def bs_tp_au(self):
        return adiabatic_bulk_modulus(self.bt_tp_au, self.alpha_tp, self.gamma_tp, self.temperature_array)

    @LazyProperty
    def bs_tp_gpa(self):
        return ry_b3_to_gpa(self.bs_tp_au)

    @LazyProperty
    def cp_tp_au(self):
        return isobaric_heat_capacity(self.cv_tp_au, self.alpha_tp, self.gamma_tp, self.temperature_array)

    @LazyProperty
    def cp_tp_jmolk(self):
        return isobaric_heat_capacity(self.cv_tp_jmolk, self.alpha_tp, self.gamma_tp, self.temperature_array)


class DifferentPhDOSCalculator(Calculator):
    def __init__(self, user_settings: Dict[str, Any]):
        super().__init__(user_settings)

        self._degeneracies = None

    @property
    def degeneracies(self):
        return self._degeneracies

    @property
    def volumes(self):
        # TODO: This is a bad style, they should be consistent
        return self._volumes[0]

    def read_input(self):
        self._degeneracies = tuple(self.settings['input'].values())
        input_data_files = tuple(self.settings['input'].keys())

        formula_unit_numbers = []
        volumes = []
        static_energies = []
        frequencies = []
        q_weights = []

        for inp in input_data_files:
            nm_tmp, volumes_tmp, static_energies_tmp, freq_tmp, weights_tmp = read_input(inp)

            if not qha.tools.is_monotonic_decreasing(volumes_tmp):
                # TODO: Clean this sentence
                save_to_output(self.settings['qha_output'], "Check the input file to make sure the volume decreases")
                raise ValueError("Check the input file to make sure the volumes are monotonic decreasing!")

            formula_unit_numbers.append(nm_tmp)
            volumes.append(volumes_tmp)
            static_energies.append(static_energies_tmp)
            frequencies.append(freq_tmp)
            q_weights.append(weights_tmp)

        formula_unit_numbers = np.array(formula_unit_numbers)
        volumes = np.array(volumes)
        static_energies = np.array(static_energies)
        frequencies = np.array(frequencies)
        q_weights = np.array(q_weights)

        if not len(set(formula_unit_numbers)) == 1:
            raise RuntimeError("All the formula unit number in all inputs should be the same!")

        if len(volumes.shape) == 1:
            raise RuntimeError("All configurations should have same number of volumes!")

        self._formula_unit_number = formula_unit_numbers[0]  # Choose any of them since they are all the same
        self._volumes = volumes
        self._static_energies = static_energies
        self._frequencies = frequencies
        self._q_weights = q_weights

    @LazyProperty
    def vib_ry(self):
        # We grep all the arguments once since they are being invoked for thousands of times, and will be an overhead.
        args = self.degeneracies, self.q_weights, self.static_energies, self._volumes, self.frequencies, \
               self.settings['static_only']

        mat = np.empty((self.temperature_array.size, self._volumes.shape[1]))
        for i, t in enumerate(self.temperature_array):
            mat[i] = different_phonon_dos.PartitionFunction(t, *(arg for arg in args)).get_free_energies()

        return mat


class SamePhDOSCalculator(DifferentPhDOSCalculator):
    def __init__(self, user_settings):
        super().__init__(user_settings)

    @LazyProperty
    def vib_ry(self):
        args = self.degeneracies, self.q_weights[0], self.static_energies, self._volumes, self.frequencies[0], \
               self.settings['static_only'], self.settings['order']
        mat = np.empty((self.temperature_array.size, self._volumes.shape[1]))

        for i, t in enumerate(self.temperature_array):
            mat[i] = same_phonon_dos.FreeEnergy(t, *(arg for arg in args)).get_free_energies()
        return mat



#!/usr/bin/env python3
"""
.. module read_input
   :platform: Unix, Windows, Mac, Linux
   :synopsis: Read file and write calculated value to files.
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
"""

import pathlib
import re
from typing import Iterator, Union, Tuple

import numpy as np
from scientific_string import strings_to_integers
from text_stream import TextStream

from qha.type_aliases import Vector, Array3D

# ===================== What can be exported? =====================
__all__ = ['read_input']


def read_input(inp: Union[str, pathlib.PurePath]) -> Tuple[int, Vector, Vector, Array3D, Vector]:
    """
    Read the standard "input" file for ``qha``.

    :param inp: The filename or its path.
    :return: The input data. They are

        - the number of formula unit in a unit cell,
        - the number of volumes in *inp*,
        - the static energies of each volume,
        - a 3D array, i.e., the frequencies of each volume of each q-point of each mode, and
        - a vector of weights of each q-point, respectively.
    """
    text_stream = TextStream(pathlib.Path(inp))
    gen: Iterator[str] = text_stream.generator_telling_position()

    volumes_amount = None
    q_points_amount = None
    modes_per_q_point_amount = None
    formula_unit_number = None
    offset = 0

    # Now we start reading some metadata.
    regex0 = re.compile("\s*(\d+)[\s,]*(\d+)[\s,]*(\d+)[\s,]*(\d+)")

    for line, offset in gen:
        if not line.strip() or line.startswith('#'):
            continue
        match = regex0.search(line)
        if match is None:
            continue
        else:
            volumes_amount, q_points_amount, modes_per_q_point_amount, formula_unit_number = strings_to_integers(
                match.groups())
            break

    # If the metadata is not found, check the *inp*!
    if any(_ is None for _ in (volumes_amount, q_points_amount, modes_per_q_point_amount)):
        raise ValueError("At least one of the desired values 'nv', 'nq', 'np' is not found in file {0}!".format(inp))

    # Generate containers for storing the following data.
    volumes = np.empty(volumes_amount, dtype=float)
    static_energies = np.empty(volumes_amount, dtype=float)
    frequencies = np.empty((volumes_amount, q_points_amount, modes_per_q_point_amount), dtype=float)
    q_weights = np.empty(q_points_amount, dtype=float)

    # We start a new iterator from where we stopped.
    gen = text_stream.generator_starts_from(offset)

    i = 0  # volume count
    j = 0  # q-point index, note it is not count like `i`!

    # Now we start reading the energies, volumes, and frequencies.
    regex1 = re.compile("P\s*=\s*-?\d*\.?\d*\s*V\s*=(\s*\d*\.?\d*)\s*E\s*=\s*(-?\d*\.?\d*)", re.IGNORECASE)

    for line in gen:
        if not line.strip():
            continue

        if '=' in line:
            match = regex1.search(line)
            if match is None:
                raise ValueError("Search of pattern {0} failed in line '{1}!".format(regex1.pattern, line))
            else:
                volumes[i], static_energies[i] = match.groups()
                i += 1
                j = 0
            continue

        sp = line.split()
        if len(sp) == 3:
            for k in range(modes_per_q_point_amount):  # Note `k` is the index of mode, like `j`, not count like `i`.
                line = next(gen)
                frequencies[i - 1, j, k] = line

            j += 1
            continue

        if 'weight' in line.lower():
            break

    # Now we start reading q-point-weights. We start the exact iterator from where we stopped.
    j = 0

    for line in gen:
        if not line.strip():
            continue

        q_weights[j] = line.split()[-1]  # This line has format: q_x q_y q_z weight, and only weight is taken
        j += 1

    if i != volumes_amount:
        raise ValueError('The number of volumes detected is not equal to what specified in head! Check your file!')

    return formula_unit_number, volumes, static_energies, frequencies, q_weights



#!/usr/bin/env python3
"""
.. module make_input
   :platform: Unix, Windows, Mac, Linux
   :synopsis: Use data files from Quantum ESPRESSO and other software to generate a standard input.
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
"""

import pathlib
import re
import warnings
from typing import Iterator, List, Tuple, Optional

import numpy as np
import yaml
from scientific_string import strings_to_integers
from text_stream import TextStream

from qha.type_aliases import Vector, Matrix

# ===================== What can be exported? =====================
__all__ = ['FromQEOutput']


class FromQEOutput:
    """
    A class that can generate a standard "input" file for the ``qha run`` command if the original data is given by
    Quantum ESPRESSO.

    :param inp_file_list: A YAML file that contains three keys and values:

        1. ``formula_unit_number``: The number of formula units in a unit cell.
        2. ``comment``: Comment that could be added as the first line of the ``input``,
           which usually specifies the system which user wants to calculate.
        3. ``frequency_files``: A YAML list of strings specifying all the frequency files given by Quantum ESPRESSO
           ``matdyn.x``.

        To become more familiar with YAML syntax, which is used in the "settings" file, please refer to
        `this documentation <http://docs.ansible.com/ansible/latest/reference_appendices/YAMLSyntax.html>`_.
    :param inp_static: A pure text file specifying the energies and pressures of each volume from the
        static calculation. The order of these volumes must be the same as the ``frequency_files`` listing order.
    :param inp_q_points: A pure text file specifying the q-points' coordinates and their corresponding weights in the
        Brillouin zone.
    """

    def __init__(self, inp_file_list: str, inp_static: str, inp_q_points: str):
        self._inp_file_list = inp_file_list
        self._inp_static = inp_static
        self._inp_q_points = inp_q_points
        self._frequency_files: Optional[List[str]] = None

        self.formula_unit_number: Optional[int] = None
        self.comment: Optional[str] = None
        self.pressures = None
        self.volumes = None
        self.static_energies = None
        self.q_coordinates = None
        self.q_weights = None
        self.frequencies = None

    def read_file_list(self) -> None:
        """
        Read all the files' names for frequency files given by Quantum ESPRESSO program ``matdyn.x``.
        """
        with open(self._inp_file_list, 'r') as f:
            d = yaml.load(f)

        self.formula_unit_number = int(d['formula_unit_number'])
        self.comment: str = d['comment']
        self._frequency_files: List[str] = d['frequency_files']

    def read_static(self) -> None:
        """
        Read static information, including pressures (which will not be used in calculation by now),
        optimized volumes, and static energies.
        """
        pressures = []
        volumes = []
        energies = []

        with open(self._inp_static, 'r') as f:
            print("Reading static data: emtpy lines or lines starting with '#' will be ignored!")

            for line in f:
                if not line.strip() or line.startswith('#'):  # Ignore empty line or comment line
                    continue

                match = re.match(r"p\s*=\s*(-?\d*\.?\d*)\s*v\s*=\s*(-?\d*\.?\d*)\s*e\s*=\s*(-?\d*\.?\d*)", line,
                                 flags=re.IGNORECASE)
                if match is None:
                    continue

                p, v, e = match.groups()
                pressures.append(p), volumes.append(v), energies.append(e)

        self.pressures = np.array(pressures, dtype=float)
        self.volumes = np.array(volumes, dtype=float)
        self.static_energies = np.array(energies, dtype=float)

    def read_q_points(self) -> None:
        """
        Read q-points' coordinates and their weights in the Brillouin zone. The q-points' coordinates are supposed to be
        three-dimensional. No other information should be given. If user still wants to,
        write lines that start with ``#``, and then they will be ignored when reading.
        """
        q_coordinates = []
        q_weights = []

        with open(self._inp_q_points, 'r') as f:
            print("Reading q-points file: emtpy lines or lines starting with '#' will be ignored!")

            regex = re.compile(r"\s*(-?\d*\.?\d*)\s*(-?\d*\.?\d*)\s*(-?\d*\.?\d*)\s*(-?\d*\.?\d*)")

            for line in f:
                if not line.strip() or line.startswith('#'):  # Ignore empty line or comment line
                    continue

                match = regex.match(line)
                if not regex.match(line):
                    raise RuntimeError(
                        "Unknown line! Should be 3 coordinates and 1 weight, other lines should be commented with '#'.")
                else:
                    g = match.groups()
                    q_coordinates.append(g[0:3])
                    q_weights.append(g[3])  # TODO: Check possible bug, what if the regex match fails

        self.q_coordinates = np.array(q_coordinates,
                                      dtype=float)  # TODO: Possible bug, ``np.array([])`` is regarded as ``False``
        self.q_weights = np.array(q_weights, dtype=float)

    @staticmethod
    def read_frequency_file(inp: str) -> Tuple[Vector, Matrix]:
        """
        Read Quantum ESPRESSO's output for phonon frequencies. This method is provided for reading only
        one file.

        :param inp: The name or path of the file.
        :return: The q-space coordinates of each q-point in the file and corresponding frequencies for each
            q-point on each band.
        """
        text_stream = TextStream(pathlib.Path(inp))

        gen: Iterator[str] = text_stream.generator_telling_position()

        q_coordinates = []
        frequencies = []

        regex = re.compile(r"nbnd\s*=\s*(\d*),\s*nks=\s*(\d*)")

        offset = None  # Initialization
        bands_amount = None
        q_points_amount = None

        for line, offset in gen:
            if not line.strip():
                continue

            if 'nbnd' or 'nks' in line:
                match = regex.search(line)

                if not match:
                    raise RuntimeError(
                        "The head line '{0}' is not complete! Here 'nbnd' and 'nks' are not found!".format(line))
                else:
                    bands_amount, q_points_amount = strings_to_integers(match.groups())
                    break

        gen: Iterator[str] = text_stream.generator_starts_from(offset)

        for line in gen:
            if not line.strip():
                continue

            q_coordinates.append(line.split())
            line = next(gen)  # Start a new line
            frequencies.append(line.split())

        q_coordinates = np.array(q_coordinates, dtype=float)
        frequencies = np.array(frequencies, dtype=float)

        if q_coordinates.shape[0] != q_points_amount:
            raise RuntimeError(
                "The number of q-points detected, {0}, is not the same as what specified in head line!".format(
                    q_coordinates.shape[0]))

        if frequencies.shape != (q_points_amount, bands_amount):
            raise RuntimeError(
                "The frequencies array shape '{0}' is not the same as what specified in head line!".format(
                    frequencies.shape))

        return q_coordinates, frequencies

    def read_frequency_files(self) -> None:
        """
        Read the phonon frequencies for all files (which are specified in the
        ``frequency_files`` key of the settings file).
        """
        frequencies_for_all_files = []

        if any(_ is None for _ in (self.q_coordinates, self.q_weights)):  # If any of them is ``None``
            self.read_q_points()  # Fill these 2 properties

        for i in range(len(self._frequency_files)):
            q_coordinates, frequencies = self.read_frequency_file(self._frequency_files[i])

            # Here I use `allclose` rather than `array_equal` since they may have very little
            # differences even they are supposed to be the same, because of the digits QE gave.
            if not np.allclose(q_coordinates, self.q_coordinates):
                warnings.warn("The q-points' coordinates are different from what specified in the q-point file!",
                              stacklevel=2)

            frequencies_for_all_files.append(frequencies)

        self.frequencies = np.array(frequencies_for_all_files)  # Shape: (# volumes, # q-points, # bands on each point)

    def write_to_file(self, outfile='input') -> None:
        """
        Write all data to a file *outfile*, which will be regarded as standard input file for ``qha``.

        :param outfile: The path or name of the output file, by default is ``'input'`` (for further calculation).
        """
        path = pathlib.Path(outfile)
        if path.is_file():
            print("Old '{0}' file found, I will backup it before continue.".format(outfile))
            path.rename(outfile + '.backup')

        with open(outfile, 'w') as f:
            f.write("# {0}\n".format(self.comment))
            f.write('# The file contains frequencies and weights at the END!\n')
            f.write('# Number of volumes (nv), q-vectors (nq), normal mode (np), formula units(nm)\n')
            # TODO: Possible bug introduced in formatting
            f.write("{0} {1} {2} {3}\n\n".format(len(self.volumes), len(self.q_weights),
                                                 self.frequencies.shape[-1], self.formula_unit_number))

            for i in range(len(self.volumes)):
                f.write("P={0:20.10f} V={1:20.10f} E={2:20.10f}\n".format(self.pressures[i], self.volumes[i],
                                                                          self.static_energies[i]))

                for j in range(len(self.q_weights)):
                    f.write("{0:12.8f} {1:12.8f} {2:12.8f}\n".format(*self.q_coordinates[j]))

                    for k in range(self.frequencies.shape[-1]):
                        f.write("{0:20.10f}\n".format(self.frequencies[i, j, k]))

            f.write('\nweight\n')
            for j in range(len(self.q_weights)):
                f.write("{0:12.8f} {1:12.8f} {2:12.8f} {3:12.8f}\n".format(*self.q_coordinates[j], self.q_weights[j]))



#!/usr/bin/env python3
"""
.. module v2p
   :platform: Unix, Windows, Mac, Linux
   :synopsis:
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

import numba
import numpy as np

from qha.tools import vectorized_find_nearest
from qha.type_aliases import Matrix, Vector

# ===================== What can be exported? =====================
__all__ = ['v2p']


@numba.jit(nopython=True, parallel=True)
def _lagrange4(x: float, x0, x1, x2, x3, y0, y1, y2, y3) -> float:
    """
    A third-order Lagrange polynomial function. Given 4 points for interpolation:
    :math:`(x_0, y_0), \ldots, (x_3, y_3)`, evaluate the Lagrange polynomial on :math:`x`.

    :param x: The x-coordinate of the point to be evaluated.
    :param x0: The x-coordinate of the 1st point.
    :param x1: The x-coordinate of the 2nd point.
    :param x2: The x-coordinate of the 3rd point.
    :param x3: The x-coordinate of the 4th point.
    :param y0: The y-coordinate of the 1st point.
    :param y1: The y-coordinate of the 2nd point.
    :param y2: The y-coordinate of the 3rd point.
    :param y3: The y-coordinate of the 4th point.
    :return: The y-coordinate of the point to be evaluated.
    """
    return (x - x1) * (x - x2) * (x - x3) / (x0 - x1) / (x0 - x2) / (x0 - x3) * y0 + \
           (x - x0) * (x - x2) * (x - x3) / (x1 - x0) / (x1 - x2) / (x1 - x3) * y1 + \
           (x - x0) * (x - x1) * (x - x3) / (x2 - x0) / (x2 - x1) / (x2 - x3) * y2 + \
           (x - x0) * (x - x1) * (x - x2) / (x3 - x0) / (x3 - x1) / (x3 - x2) * y3


def v2p(func_of_t_v: Matrix, p_of_t_v: Matrix, desired_pressures: Vector) -> Matrix:
    """
    Obtain :math:`f(T, P)` given :math:`f(T, V)` and :math:`P(T, V)` by doing a fourth-order Lagrangian interpolation.

    :param func_of_t_v: Any function :math:`f` on :math:`(T, V)` grid, which has
        shape: (number of temperature, number of volumes).
    :param p_of_t_v: Pressures on :math:`(T, V)` grid, which has
        shape: (number of temperature, number of volumes).
    :param desired_pressures: A vector of pressures which user wants to apply.
    :return: The interpolated function :math:`f` on :math:`(T, P)` grid.
    """
    t_amount, v_amount = func_of_t_v.shape
    result = np.empty((t_amount, v_amount))

    extended_f = np.hstack((func_of_t_v[:, 3].reshape(-1, 1), func_of_t_v, func_of_t_v[:, -4].reshape(-1, 1)))
    extended_p = np.hstack((p_of_t_v[:, 3].reshape(-1, 1), p_of_t_v, p_of_t_v[:, -4].reshape(-1, 1)))

    desired_pressures_amount = len(desired_pressures)
    for i in range(t_amount):
        rs = np.zeros(desired_pressures_amount)
        vectorized_find_nearest(extended_p[i], desired_pressures, rs)

        for j in range(v_amount):
            k = int(rs[j])
            x1, x2, x3, x4 = extended_p[i, k - 1:k + 3]
            f1, f2, f3, f4 = extended_f[i, k - 1:k + 3]

            result[i, j] = _lagrange4(desired_pressures[j], x1, x2, x3, x4, f1, f2, f3, f4)
    return result



#!/usr/bin/env python3
"""
.. module thermodynamics
   :platform: Unix, Windows, Mac, Linux
   :synopsis: Calculate the basic thermodynamic properties for a system: S, U, H, G, Cv on a (T,V) grid.
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

import numpy as np

from qha.type_aliases import Matrix, Vector
from qha.v2p import v2p

# ===================== What can be exported? =====================
__all__ = [
    'pressure',
    'entropy',
    'thermodynamic_potentials',
    'volume',
    'thermal_expansion_coefficient',
    'gruneisen_parameter',
    'isothermal_bulk_modulus',
    'adiabatic_bulk_modulus',
    'bulk_modulus_derivative',
    'isobaric_heat_capacity',
    'volumetric_heat_capacity'
]


def calculate_derivatives(xs: Vector, fs: Matrix) -> Matrix:
    """
    Calculate the derivative of :math:`f(x)`, i.e., :math:`\\frac{ df(x) }{ dx }`.

    :param xs: A 1D vector, with length :math:`N_x`.
    :param fs: A matrix, with shape :math:`(N_x, _)`.
    :return: A matrix, with shape :math:`(N_x, _)`.
    """
    if xs.ndim > 1 or fs.ndim < 2:
        raise ValueError('The argument *xs* should be a 1D array and *ys* should be a 2D array!')

    return np.gradient(fs, axis=0) / np.gradient(xs)[:, None]  # df(x)/dx.


def pressure(vs: Vector, free_energies: Matrix) -> Matrix:
    """
    Calculate the pressure as a function of temperature and volume, i.e.,

    .. math::

       P = - \\bigg( \\frac{ \\partial F(T, V) }{ \\partial V } \\bigg)_T.

    :param vs: A vector of volumes.
    :param free_energies: A matrix, the free energy as a function of temperature and volume, i.e., :math:`F(T, V)`.
    :return: A matrix, the pressure as a function of temperature and volume, i.e., :math:`P(T,V)`.
    """
    return -np.gradient(free_energies, axis=1) / np.gradient(vs)


def entropy(temperature: Vector, free_energies: Matrix) -> Matrix:
    """
    Calculate the entropy as a function of temperature and volume, i.e.,

    .. math::

       S = - \\bigg( \\frac{ \\partial F(T, V) }{ \\partial T } \\bigg)_V.

    :param temperature: A vector of temperature.
    :param free_energies: A matrix, the free energy as a function of temperature and volume, i.e., :math:`F(T, V)`.
    :return: A matrix, the entropy as a function of temperature and volume, i.e., :math:`S(T,V)`.
    """
    return -calculate_derivatives(temperature, free_energies)


def thermodynamic_potentials(temperature: Vector, vs: Vector, free_energies: Matrix, ps: Matrix):
    """
    Calculate the enthalpy :math:`H(T, V)`, the internal energy :math:`U(T, V)`,
    and the Gibbs free energy :math:`G` on a :math:`(T, V)` grid from Helmholtz free energy :math:`F(T, V)` by

    .. math::

       U(T, V) &= F(T, V) + T S(T, V), \\\\
       H(T, V) &= U(T, V) + P(T, V) V, \\\\
       G(T, V) &= F(T, V) + P(T, V) V.

    :param temperature: A vector of temperature.
    :param vs: A vector of volumes.
    :param free_energies: A matrix, the free energy as a function of temperature and volume, i.e., :math:`F(T, V)`.
    :param ps: A matrix, the pressure as a function of temperature and volume, i.e., :math:`P(T, V)`.
    :return: A dictionary that contains the enthalpy :math:`H(T, V)`, the internal energy :math:`U(T, V)`,
        and the Gibbs free energy :math:`G` on a :math:`(T, V)` grid. They can be retrieved by ``'U'``, ``'H'``, or
        ``'G'`` keys, respectively.
    """
    g: Matrix = free_energies + ps * vs  # G(T,V) = F(T,V) + V * P(T,V)

    u: Matrix = free_energies + entropy(temperature, free_energies) * temperature.reshape(-1, 1)  # U(T,V) = F(T,V) + T * S(T,V)

    h: Matrix = u + ps * vs  # H(T,V) = U(T,V) + V * P(T,V)

    return {'U': u, 'H': h, 'G': g}


def volume(vs: Vector, desired_ps: Vector, ps: Matrix) -> Matrix:
    """
    Convert the volumes as a function of temperature and pressure, i.e., on a :math:`(T, P)` grid.

    :param vs: A vector of volumes.
    :param desired_ps: A vector of desired pressures.
    :param ps: A matrix, the pressure as a function of temperature and volume, i.e., :math:`P(T,V)`, in atomic unit.
    :return: A matrix, the volume as a function of temperature and pressure, i.e., :math:`V(T, P)`.
    """
    nt, ntv = ps.shape
    vs = vs.reshape(1, -1).repeat(nt, axis=0)
    return v2p(vs, ps, desired_ps)


def thermal_expansion_coefficient(temperature: Vector, vs: Matrix) -> Matrix:
    """
    Calculate the thermal expansion coefficient by

    .. math::

       \\alpha = \\frac{ 1 }{ V }  \\bigg( \\frac{ \\partial V }{ \\partial T } \\bigg)_P.

    :param temperature: A vector of temperature.
    :param vs: A matrix, the volume as a function of temperature and pressure, i.e., :math:`V(T, P)`.
    :return: A matrix, the thermal expansion coefficient as a function of temperature and pressure,
        i.e., :math:`\\alpha(T, P)`.
    """
    # Division is done by element-wise.
    return calculate_derivatives(temperature, vs) / vs


def gruneisen_parameter(vs: Matrix, bt: Matrix, alpha: Matrix, cv: Matrix) -> Matrix:
    """
    Calculate the Grüneisen parameter by

    .. math::

       \\gamma = \\frac{ \\alpha B_T V }{ C_V }.

    :param vs: A matrix, the volume as a function of temperature and pressure, i.e., :math:`V(T, P)`.
    :param bt: A matrix, the isothermal bulk modulus as a function of temperature and pressure,
        i.e., :math:`B_T(T, P)`.
    :param alpha: A matrix, the thermal expansion coefficient as a function of temperature and pressure,
        i.e., :math:`\\alpha(T, P)`.
    :param cv: A matrix, the volumetric heat capacity as a function of temperature and pressure,
        i.e., :math:`C_V(T, P)`.
    :return: A matrix, the thermodynamic Grüneisen parameter as a function of temperature and pressure,
        i.e., :math:`\\gamma(T, P)`.
    """
    gamma = np.empty([vs.shape[0], vs.shape[1]])
    gamma[0] = 0.0
    gamma[1:, :] = vs[1:] * bt[1:] * alpha[1:] / cv[1:]
    # return vs * bt * alpha / cv
    return gamma


def isothermal_bulk_modulus(vs: Vector, ps: Matrix) -> Matrix:
    """
    Calculate the isothermal bulk modulus by

    .. math::

       B_T = - V \\bigg( \\frac{ \\partial P }{ \\partial V } \\bigg)_T.

    :param vs: A vector of volumes.
    :param ps: A matrix, the pressure as a function of temperature and volume, i.e., :math:`P(T, V)`.
    :return: A matrix, the isothermal bulk modulus, as a function of temperature and volume, i.e., :math:`B_T(T, V)`.
    """
    return -np.gradient(ps, axis=1) / np.gradient(vs) * vs


def adiabatic_bulk_modulus(bt: Matrix, alpha: Matrix, gamma: Matrix, temperature: Vector) -> Matrix:
    """
    Calculate the adiabatic bulk modulus by

    .. math::

       B_S = B_T \\big( 1 + \\alpha \\gamma T \\big).

    :param bt: A matrix, the isothermal bulk modulus, as a function of temperature and pressure,
        i.e., :math:`B_T(T, P)`.
    :param alpha: A matrix, the thermal expansion coefficient as a function of temperature and pressure,
        i.e., :math:`\\alpha(T, P)`.
    :param gamma: A matrix, the thermodynamic Grüneisen parameter as a function of temperature and pressure,
        i.e., :math:`\\gamma(T, P)`.
    :param temperature: A vector of temperature.
    :return: A matrix, the adiabatic bulk modulus, as a function of temperature and pressure,
        i.e., :math:`B_S(T,P)`.
    """
    return bt * (1.0 + alpha * gamma * temperature[:, None])


def bulk_modulus_derivative(ps: Vector, bt: Matrix) -> Matrix:
    """
    Calculate the first-order derivative of bulk modulus with respect to pressure by

    .. math::

       B_T' = \\bigg( \\frac{ \\partial B_T }{ \\partial P } \\bigg).

    :param ps: A vector of pressures.
    :param bt: A matrix, the isothermal bulk modulus, as a function of temperature and pressure,
        i.e., :math:`B_T(T, P)`.
    :return: A matrix, the isothermal bulk modulus, as a function of temperature and pressure, i.e.,
        :math:`B_T'(T, P)`.
    """
    return calculate_derivatives(ps, bt.T).T


def isobaric_heat_capacity(cv: Matrix, alpha: Matrix, gamma: Matrix, temperature: Vector) -> Matrix:
    """
    Calculate the isobaric heat capacity by

    .. math::

       C_P = C_V \\big( 1 + \\alpha \\gamma T \\big).

    :param cv: A matrix, the volumetric heat capacity, :math:`C_V(T, P)`.
    :param alpha: A matrix, the thermal expansion coefficient as a function of temperature and pressure,
        i.e., :math:`\\alpha(T, P)`.
    :param gamma: A matrix, the thermodynamic Grüneisen parameter as a function of temperature and pressure,
        i.e., :math:`\\gamma(T, P)`.
    :param temperature: A vector of temperature.
    :return: A matrix, the isobaric specific heat capacity as a function of temperature and pressure,
        i.e., :math:`C_P(T,P)`.
    """
    return cv * (1.0 + alpha * gamma * temperature[:, None])


def volumetric_heat_capacity(temperature: Vector, internal_energies: Matrix) -> Matrix:
    """
    Calculate the volumetric heat capacity by

    .. math::

       C_V = \\bigg( \\frac{ \\partial U }{ \\partial T } \\bigg).

    :param temperature: A vector of temperature.
    :param internal_energies: A matrix, the internal energy as a function of temperature and volume,
        i.e., :math:`U(T, V)`.
    :return: A matrix, the volumetric heat capacity as a function of temperature and volume,
        i.e., :math:`C_V(T, V)`.
    """
    return calculate_derivatives(temperature, internal_energies)



#!/usr/bin/env python3
"""
.. module tools
   :platform: Unix, Windows, Mac, Linux
   :synopsis: This module defines some tools used in this package.
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
"""

from typing import Callable, Optional

import numpy as np
from numba import float64, guvectorize, int64, jit, vectorize

from qha.fitting import polynomial_least_square_fitting
from qha.grid_interpolation import calculate_eulerian_strain
from qha.type_aliases import Matrix, Scalar, Vector

# ===================== What can be exported? =====================
__all__ = ['find_nearest', 'vectorized_find_nearest', 'lagrange3', 'lagrange4', 'is_monotonic_decreasing',
           'is_monotonic_increasing', 'arange', 'calibrate_energy_on_reference']


def lagrange4(xs: Vector, ys: Vector) -> Callable[[float], float]:
    """
    A third-order Lagrange polynomial function. Given 4 points for interpolation:
    :math:`(x_0, y_0), \ldots, (x_3, y_3)`, evaluate the Lagrange polynomial on :math:`x`, referenced from
    `Wolfram MathWorld <http://mathworld.wolfram.com/LagrangeInterpolatingPolynomial.html>`_.

    :param xs: A vector of the x-coordinates' of the 4 points.
    :param ys: A vector of the y-coordinates' of the 4 points.
    :return: A function that can evaluate the value of an x-coordinate within the range of
        :math:`[\\text{min}(xs), \\text{max}(xs)]`, where :math:`xs` denotes the argument *xs*.
    """
    if not len(xs) == len(ys) == 4:
        raise ValueError('The *xs* and *ys* must both have length 4!')

    x0, x1, x2, x3 = xs
    y0, y1, y2, y3 = ys

    @vectorize(["float64(float64)"], target='parallel')
    def f(x: float) -> float:
        """
        A helper function that only does the evaluation.

        :param x: The variable on which the Lagrange polynomial is going to be applied.
        :return: The value of the Lagrange polynomial on :math:`x`, i.e., :math:`L(x)`.
        """
        return (x - x1) * (x - x2) * (x - x3) / (x0 - x1) / (x0 - x2) / (x0 - x3) * y0 + \
               (x - x0) * (x - x2) * (x - x3) / (x1 - x0) / (x1 - x2) / (x1 - x3) * y1 + \
               (x - x0) * (x - x1) * (x - x3) / (x2 - x0) / (x2 - x1) / (x2 - x3) * y2 + \
               (x - x0) * (x - x1) * (x - x2) / (x3 - x0) / (x3 - x1) / (x3 - x2) * y3

    return f


def lagrange3(xs: Vector, ys: Vector) -> Callable[[float], float]:
    """
    A second-order Lagrange polynomial function. Given 3 points for interpolation:
    :math:`(x_0, y_0), \ldots, (x_2, y_2)`, evaluate the Lagrange polynomial on :math:`x`, referenced from
    `Wolfram MathWorld also <http://mathworld.wolfram.com/LagrangeInterpolatingPolynomial.html>`_.

    .. doctest::

        >>> xs = [0, 1, 3]
        >>> ys = [2, 4, 5]
        >>> f = lagrange3(xs, ys)
        >>> f(2.5)
        5.125

    :param xs: A vector of the x-coordinates' of the 3 points.
    :param ys: A vector of the y-coordinates' of the 3 points.
    :return: A function that can evaluate the value of an x-coordinate within the range of
        :math:`[\\text{min}(xs), \\text{max}(xs)]`, where :math:`xs` denotes the argument *xs*.
    """
    if not len(xs) == len(ys) == 3:
        raise ValueError('The *xs* and *ys* must both have length 3!')

    if len(set(xs)) < 3:  # The ``set`` will remove duplicated items.
        raise ValueError('Some elements of *xs* are duplicated!')

    x0, x1, x2 = xs
    y0, y1, y2 = ys

    @vectorize(["float64(float64)"], target='parallel')
    def f(x: float) -> float:
        """
        A helper function that only does the evaluation.

        :param x: The variable on which the Lagrange polynomial is going to be applied.
        :return: The value of the Lagrange polynomial on :math:`x`, i.e., :math:`L(x)`.
        """
        return (x - x1) * (x - x2) / (x0 - x1) / (x0 - x2) * y0 + \
               (x - x0) * (x - x2) / (x1 - x0) / (x1 - x2) * y1 + \
               (x - x0) * (x - x1) / (x2 - x0) / (x2 - x1) * y2

    return f


@jit(nopython=True, nogil=True, cache=True)
def find_nearest(array: Vector, value: Scalar) -> int:
    """
    Given an *array* , and given a *value* , returns an index ``j`` such that *value* is between ``array[j]``
    and ``array[j+1]``. The *array* must be monotonic increasing. ``j=-1`` or ``j=len(array)`` is returned
    to indicate that *value* is out of range below and above respectively.
    If *array* is unsorted, consider first using an :math:`O(n \log n)` sort and then use this function.
    Referenced from `Stack Overflow <https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array>`_.

    :param array: An array of monotonic increasing real numbers.
    :param value: The value which user wants to find between two of the consecutive elements in *array*.
    :return: The index mentioned above.

    .. doctest::

        >>> find_nearest([1, 3, 4, 6, 9, 11], 3.5)
        1
        >>> find_nearest([1, 3, 4, 6, 9, 11], 0)
        0
        >>> find_nearest([1, 3, 4, 6, 9, 11], 12)
        4
    """
    n: int = len(array)

    if value <= array[0]:
        return 0
    elif value >= array[-1]:
        return n - 2

    j_low = 0  # Initialize lower limit.
    j_up = n - 1  # Initialize upper limit.

    while j_up - j_low > 1:  # If we are not yet done,
        j_mid: int = (j_up + j_low) // 2  # compute a midpoint,
        if value >= array[j_mid]:
            j_low = j_mid  # and replace either the lower limit
        else:
            j_up = j_mid  # or the upper limit, as appropriate.
        # Repeat until the test condition is satisfied.

    return j_low


@guvectorize([(float64[:], float64[:], int64[:])], '(m),(n)->(n)')
def vectorized_find_nearest(array: Vector, values: Vector, result: Vector):
    """
    A vectorized version of function ``find_nearest``.

    :param array: An array of monotonic increasing real numbers.
    :param values: An array of values, each of which is one between two of the consecutive elements in *array*.
    :param result: An array of indices. It is suggested to generate a vector of zeros by ``numpy`` package.
    :return: The *result*, an array of indices mentioned above.
    """
    n: int = len(array)

    if len(values) != len(result):
        raise ValueError('The *values* and *result* arguments should have same length!')

    for i in range(len(values)):

        if values[i] <= array[0]:
            result[i] = 0
        elif values[i] >= array[-1]:
            result[i] = n - 2

        j_low = 0  # Initialize lower limit.
        j_up = n - 1  # Initialize upper limit.

        while j_up - j_low > 1:  # If we are not yet done,
            j_mid: int = (j_up + j_low) // 2  # compute a midpoint,
            if values[i] >= array[j_mid]:
                j_low = j_mid  # and replace either the lower limit
            else:
                j_up = j_mid  # or the upper limit, as appropriate.
            # Repeat until the test condition is satisfied.

        result[i] = j_low


def arange(start: Scalar, num: int, step: Scalar) -> Vector:
    """
    Similar to ``numpy.arange``, generate an one-dimensional array but with *start*, *num*, *step* specified.

    :param start: The starting point of the array, which is included in this array.
    :param num: The total number of elements in this array.
    :param step: A number specifying the difference between any of the two elements.
    :return: An arithmetic progression.
    """
    return np.array([start + step * n for n in range(int(num))])


def is_monotonic_decreasing(array: Vector) -> bool:
    """
    Check whether the *array* is monotonic decreasing or not.
    For example, in QHA calculation, the volumes should be listed as a decreasing array,
    while the pressures should be monotonic increasing.
    This function can be used to check whether the volumes are in the right order.

    .. doctest::

        >>> is_monotonic_decreasing([1, 2, 4, 5, 9])
        False
        >>> is_monotonic_decreasing([2, -5, -10, -20])
        True

    :param array: The array to be evaluated.
    :return: ``True`` if the argument *array* is monotonic decreasing, otherwise ``False``.
    """
    dx = np.diff(array)
    return np.all(dx <= 0)


def is_monotonic_increasing(array: Vector) -> bool:
    """
    Check whether the *array* is monotonic increasing or not.
    For example, in QHA calculation, the volumes should be listed as decreasing array,
    and the pressures should be monotonic increasing.
    This function can be used to check whether the pressures are in the right order.

    .. doctest::

        >>> is_monotonic_increasing([1, 2, 4, 5, 9])
        True
        >>> is_monotonic_increasing([2, -5, -10, -20])
        False

    :param array: The array to be evaluated.
    :return: ``True`` if the argument *array* is monotonic increasing, otherwise ``False``.
    """
    dx = np.diff(array)
    return np.all(dx >= 0)


def calibrate_energy_on_reference(volumes_before_calibration: Matrix, energies_before_calibration: Matrix,
                                  order: Optional[int] = 3):
    """
    In multi-configuration system calculation, the volume set of each calculation may vary a little,
    This function would make the volume set of the first configuration (normally, the most populated configuration)
    as a reference volume set, then calibrate the energies of all configurations to this reference volume set.

    :param volumes_before_calibration: Original volume sets of all configurations
    :param energies_before_calibration: Free energies of all configurations on the original volume sets.
    :param order: The order of Birch--Murnaghan EOS fitting.
    :return: Free energies of each configuration on the
        referenced volumes (usually the volumes of the first configuration).
    """
    configurations_amount, _ = volumes_before_calibration.shape
    volumes_for_reference: Vector = volumes_before_calibration[0]

    energies_after_calibration = np.empty(volumes_before_calibration.shape)
    for i in range(configurations_amount):
        strains_before_calibration = calculate_eulerian_strain(volumes_before_calibration[i, 0],
                                                               volumes_before_calibration[i])
        strains_after_calibration = calculate_eulerian_strain(volumes_before_calibration[i, 0], volumes_for_reference)
        _, energies_after_calibration[i, :] = polynomial_least_square_fitting(strains_before_calibration,
                                                                              energies_before_calibration[i],
                                                                              strains_after_calibration,
                                                                              order=order)
    return energies_after_calibration


if __name__ == '__main__':
    import doctest

    doctest.testmod()



#!/usr/bin/env python3
"""
.. module multi_configurations.same_phonon_dos
   :platform: Unix, Windows, Mac, Linux
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

from typing import Optional

import numpy as np
from lazy_property import LazyProperty
from scipy.constants import physical_constants as pc
from scipy.special import logsumexp

import qha.settings
from qha.statmech import ho_free_energy, log_subsystem_partition_function
from qha.tools import calibrate_energy_on_reference
from qha.type_aliases import Array3D, Scalar, Vector, Matrix

# ===================== What can be exported? =====================
__all__ = ['PartitionFunction', 'FreeEnergy']

K = {'ha': pc['Boltzmann constant in eV/K'][0] / pc['Hartree energy in eV'][0],
     'ry': pc['Boltzmann constant in eV/K'][0] / pc['Rydberg constant times hc in eV'][0],
     'ev': pc['Boltzmann constant in eV/K'][0],
     'SI': pc['Boltzmann constant'][0]}[qha.settings.energy_unit]


class PartitionFunction:
    """
    A class that represents the partition function of multiple configurations with the same phonon density of states.
    In mathematics, it is represented as

    .. math::

       Z_{\\text{all configs}}(T, V) = \sum_{j = 1}^{N_{c}} g_{j}
       \\bigg\{
         \exp \\Big( -\\frac{ E_j(V) }{ k_B T } \\Big)
         \prod_{\mathbf{q}s} \\bigg(
           \\tfrac{\exp \\big(-\\tfrac{ \hbar \omega_{\mathbf{ q }s}^j(V) }{ 2 k_B T }\\big)}{1 - \exp \\big(-\\tfrac{ \hbar \omega_{\mathbf{ q }s}^j(V) }{ k_B T }\\big)}
         \\bigg)^{w_{\mathbf{ q }}^j}
       \\bigg\}.

    :param temperature: The temperature at which the partition function is calculated.
    :param degeneracies: An array of degeneracies of each configuration, which will not be normalized in the
        calculation. They should all be positive integers.
    :param q_weights: The weights of all q-points that are sampled, should be a vector since all configurations should
        have the same q-point weights.
    :param static_energies: The static energies of each configuration of each volume.
    :param frequencies: It is a 3D array that specifies the frequency on each volume, q-point and mode.
        It is not 4D since we have all configurations sharing the same phonon density of states.
    :param precision: The precision of a big float number to represent the partition function since it is a very large
        value, by default, is ``500``.
    """

    def __init__(self, temperature: Scalar, degeneracies: Vector, q_weights: Vector, static_energies: Matrix,
                 frequencies: Array3D, precision: Optional[int] = 500):

        if not np.all(np.greater_equal(degeneracies, 0)):
            raise ValueError('Degeneracies should all be integers greater equal than 0!')
        if not np.all(np.greater_equal(q_weights,
                                       0)):  # Weights should all be greater equal than 0, otherwise sum will be wrong.
            raise ValueError('Weights should all be greater equal than 0!')

        self.frequencies = np.array(frequencies)
        if self.frequencies.ndim != 3:
            raise ValueError("*frequencies* must be a three-dimensional array!")

        if temperature < 1e-6:
            self.temperature = 1e-6
        else:
            self.temperature = temperature
        self.static_energies = np.array(static_energies)
        self.degeneracies = np.array(degeneracies)
        self.q_weights = np.array(q_weights)
        self._scaled_q_weights = self.q_weights / np.sum(q_weights)
        self.precision = int(precision)

    @property
    def _static_part(self) -> Vector:
        """
        Calculate the static contribution to the partition function.

        :return: The static contribution on the temperature-volume grid.
        """
        try:
            import bigfloat
        except ImportError:
            raise ImportError(
                "You need to install ``bigfloat`` package to use {0} object!".format(self.__class__.__name__))

        with bigfloat.precision(self.precision):
            return np.array([bigfloat.exp(d) for d in  # shape = (# of volumes for each configuration, 1)
                             logsumexp(-self.static_energies / (K * self.temperature), axis=1, b=self.degeneracies)])

    @property
    def _harmonic_part(self) -> Vector:
        """
        Calculate the harmonic contribution to the partition function.

        :return: The harmonic contribution on the temperature-volume grid.
        """
        log_product_modes: Matrix = np.sum(
            log_subsystem_partition_function(self.temperature, self.frequencies), axis=2, dtype=float)
        return np.exp(np.dot(log_product_modes, self._scaled_q_weights))  # (vol, 1)

    @LazyProperty
    def total(self) -> Vector:
        """
        Calculate the total partition function.

        :return: The partition function on the temperature-volume grid.
        """
        return np.multiply(self._static_part, self._harmonic_part)  # (vol, 1), product element-wise

    def get_free_energies(self):
        """
        Give the free energy by

        .. math::

           F_{\\text{all configs}}(T, V) = - k_B T \ln Z_{\\text{all configs}}(T, V).

        :return: The free energy on the temperature-volume grid.
        """
        try:
            import bigfloat
        except ImportError:
            raise ImportError("Install ``bigfloat`` package to use {0} object!".format(self.__class__.__name__))

        with bigfloat.precision(self.precision):
            log_z = np.array([bigfloat.log(d) for d in self.total], dtype=float)
        return -K * self.temperature * log_z


class FreeEnergy:
    """
    A class that represents the free energy of multiple configurations with the same phonon density of states.
    In mathematics, it is represented as

    .. math::

       F_{\\text{all configs}}(T, V) = - k_B T \ln Z_{\\text{all configs}}(T, V)
       = - k_B T \ln \\bigg( \sum_{j = 1}^{N_{c}} g_{j} \exp \\Big( -\\frac{ E_j(V) }{ k_B T } \\Big) \\bigg)
         + \sum_{\mathbf{ q }s} w_\mathbf{ q }
             \\bigg\{ \\frac{ \hbar \omega_{\mathbf{ q }s}(V) }{ 2 }
             + k_B \ln \\bigg( 1 - \exp \\Big( -\\frac{ \hbar \omega_{\mathbf{ q }s}(V) }{ k_B T } \\Big) \\bigg)
         \\bigg\}.

    :param temperature: The temperature at which the partition function is calculated.
    :param degeneracies: An array of degeneracies of each configuration, which will not be normalized in the
        calculation. They should all be positive integers.
    :param q_weights: The weights of all q-points that are sampled, should be a vector since all configurations should
        have the same q-point weights.
    :param static_energies: The static energy of each configuration of each volume.
    :param volumes: A matrix of volumes of each configurations, should have the same values for each
        configuration.
    :param frequencies: It is a 3D array that specifies the frequency on each volume, q-point and mode.
        It is not 4D since we have all configurations sharing the same phonon density of states.
    :param static_only: Whether the calculation only takes static contribution and does not consider
        the vibrational contribution, by default, is ``False``.
    :param order: The order of Birch--Murnaghan equation of state fitting, by default, is ``3``.
    """

    def __init__(self, temperature: Scalar, degeneracies: Vector, q_weights: Vector, static_energies: Matrix,
                 volumes: Matrix, frequencies: Array3D, static_only: Optional[bool] = False, order: Optional[int] = 3):
        if not np.all(np.greater_equal(degeneracies, 0)):
            raise ValueError('Degeneracies should all be integers greater equal than 0!')
        if not np.all(np.greater_equal(q_weights,
                                       0)):  # Weights should all be greater equal than 0, otherwise sum will be wrong.
            raise ValueError('Weights should all be greater equal than 0!')

        self.frequencies = np.array(frequencies)
        if self.frequencies.ndim != 3:
            raise ValueError("*frequencies* must be a three-dimensional array!")

        if temperature < 1e-6:
            self.temperature = 1e-6
        else:
            self.temperature = temperature
        self.static_energies = np.array(static_energies)
        self.volumes = np.array(volumes)
        self.degeneracies = np.array(degeneracies)
        self.q_weights = np.array(q_weights)
        self._scaled_q_weights = self.q_weights / np.sum(q_weights)
        self.static_only = static_only
        self.order = order

    @LazyProperty
    def aligned_static_energies_for_each_configuration(self):
        """
        If the input static energies are not aligned for each configuration, then do a
        fitting to align all static energies.

        :return: A matrix of aligned static energy of each configuration of each volume.
        """
        return calibrate_energy_on_reference(self.volumes, self.static_energies, self.order)

    @LazyProperty
    def static_part(self) -> Vector:
        """
        Calculate the static contribution to the free energy.

        :return: The static contribution on the temperature-volume grid.
        """
        kt: float = K * self.temperature  # k_B T
        inside_exp: Matrix = -self.aligned_static_energies_for_each_configuration.T / kt  # exp( E_n(V) / k_B / T )
        return -kt * logsumexp(inside_exp, axis=1, b=self.degeneracies)

    @LazyProperty
    def harmonic_part(self) -> Vector:
        """
        Calculate the harmonic contribution to the free energy.

        :return: The harmonic contribution on the temperature-volume grid.
        """
        sum_modes = np.sum(ho_free_energy(self.temperature, self.frequencies), axis=2)
        return np.dot(sum_modes, self._scaled_q_weights)

    @LazyProperty
    def total(self) -> Vector:
        """
        Calculate the total partition function, which is the static part plus the harmonic part.

        :return: Total free energy on the temperature-volume grid.
        """
        return self.static_part + self.harmonic_part

    def get_free_energies(self):
        """
        If ``static_only = True`` is specified in class instantiation, then only the static contribution will
        be counted. Then it is equivalent to the ``static_part`` property. If not, then this is equivalent to
        the ``total`` property.

        :return: The free energy on the temperature-volume grid.
        """
        if self.static_only:
            return self.static_part

        return self.total



#!/usr/bin/env python3
"""
.. module type_aliases
   :platform: Unix, Windows, Mac, Linux
   :synopsis: This module defines several data types that will remind the user what input should be given
    throughout the calculation.
    The ``Scalar`` type is just a number type;
    the ``Vector`` type is a 1D-array type, can be a list, a tuple, or a ``numpy`` array of floating-point numbers;
    the ``Matrix`` type is a 2D-array type, containing two dimensions of floating-point numbers;
    the ``Array3D`` type is a 3D-array type, containing three dimensions of floating-point numbers;
    and the ``Array4D`` type is a 4D-array type, containing four dimensions of floating-point numbers.
    These types will not affect the code, they are defined to tell the users what is inside the developers' minds
    when they are coding.
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

from numba import float64

# ===================== What can be exported? =====================
__all__ = ['Scalar', 'Vector', 'Matrix', 'Array3D', 'Array4D']

Scalar = float64  # 0-dimensional float
Vector = float64[:]  # 1-dimensional floats
Matrix = float64[:, :]  # 2-dimensional floats
Array3D = float64[:, :, :]  # 3-dimensional floats
Array4D = float64[:, :, :, :]  # 4-dimensional floats



#!/usr/bin/env python3
"""
.. module cli.converter
   :platform: Unix, Windows, Mac, Linux
.. moduleauthor:: Chenxing Luo <chenxing.luo@columbia.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

from qha.cli.handler import QHACommandHandler
from qha.basic_io.input_maker import FromQEOutput


class QHAConverter(QHACommandHandler):
    def __init__(self):
        super().__init__()

    def init_parser(self, parser):
        super().init_parser(parser)
        parser.add_argument('inp_file_list', type=str)
        parser.add_argument('inp_static', type=str)
        parser.add_argument('inp_q_points', type=str)

    def run(self, namespace):
        inp_file_list = namespace.inp_file_list
        inp_static = namespace.inp_static
        inp_q_points = namespace.inp_q_points

        converter = FromQEOutput(inp_file_list, inp_static, inp_q_points)
        converter.read_file_list()
        converter.read_static()
        converter.read_q_points()
        converter.read_frequency_files()
        converter.write_to_file()



#!/usr/bin/env python3
"""
.. module cli
   :platform: Unix, Windows, Mac, Linux
.. moduleauthor:: Chenxing Luo <chenxing.luo@columbia.edu>
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
"""

from .converter import QHAConverter
from .parser import QHAArgumentParser
from .plotter import QHAPlotter
from .runner import QHARunner


def main():
    parser = QHAArgumentParser()

    qha_converter = QHAConverter()
    parser.register_handler('convert', qha_converter, 'conv')

    qha_runner = QHARunner()
    parser.register_handler('run', qha_runner)

    qha_plotter = QHAPlotter()
    parser.register_handler('plot', qha_plotter)

    parser.load_plugins()

    namespace = parser.parse_args()
    parser.invoke_handler(namespace)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
.. module
   :platform: Unix, Windows, Mac, Linux
   :synopsis:
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""
import matplotlib

matplotlib.use('Agg')
import matplotlib.pylab as plt
import pandas as pd
import seaborn as sns
import pathlib
import numpy as np

# ===================== What can be exported? =====================
__all__ = ['Plotter']


class Plotter:
    def __init__(self, user_settings):
        self.user_settings = user_settings
        self.results_folder = pathlib.Path(self.user_settings['output_directory'])

    def plot2file(self, plot_what):
        if plot_what == 'alpha':
            self.plot_thermal_expansion()
        if plot_what == 'gamma':
            self.plot_gruneisen_parameter()
        if plot_what == 'Cp':
            self.plot_pressure_specific_heat_capacity()
        if plot_what == 'Bt':
            self.plot_isothermal_bulk_modulus()
        if plot_what == 'G':
            self.plot_gibbs_free_energy()
        if plot_what == 'V':
            self.plot_volume()

    @staticmethod
    def plot_to_file():
        size_small = 12
        size_medium = 14
        size_big = 16
        fontweight = 'normal'  # normal, bold, bolder, lighter

        # Defind the font size
        plt.rcParams['font.size'] = size_small
        plt.rcParams['font.weight'] = fontweight
        plt.rcParams['axes.titlesize'] = size_medium
        plt.rcParams['axes.labelsize'] = size_medium
        plt.rcParams['legend.fontsize'] = size_small
        plt.rcParams['figure.titlesize'] = size_big

        plt.rcParams['axes.linewidth'] = 1.5
        plt.rcParams['xtick.major.size'] = 6
        plt.rcParams['xtick.minor.size'] = 3
        plt.rcParams['ytick.major.size'] = 6
        plt.rcParams['ytick.minor.size'] = 3
        plt.rcParams['lines.linewidth'] = 2

        # Define the x and y tick features
        plt.rcParams['xtick.labelsize'] = size_small
        plt.rcParams['ytick.labelsize'] = size_small
        plt.rcParams['xtick.direction'] = 'in'
        plt.rcParams['ytick.direction'] = 'in'
        plt.rcParams['xtick.minor.visible'] = True
        plt.rcParams['ytick.minor.visible'] = True
        # Define the default figure size in inches
        plt.rcParams['figure.figsize'] = [8, 6]
        # Define Figures overall properties
        plt.rcParams['figure.dpi'] = 300

        plt.figure()

    @staticmethod
    def save_fig(ax, fig_name):
        fig = ax.get_figure()
        fig_name_pdf = str(fig_name) + '.pdf'
        fig.savefig(fig_name_pdf, format='pdf', dpi=300, bbox_inches='tight')

    def plot_thermal_expansion(self):
        f_alpha = self.user_settings['alpha']
        alpha = pd.read_csv(f_alpha, sep='\s+', index_col='T(K)\P(GPa)')
        self.plot_to_file()
        ax = alpha.plot(figsize=(6, 5), color=sns.color_palette("Reds", len(alpha.columns)))
        ax.set_xlabel('Temperature (K)')
        ax.set_ylabel('Thermal Expansion')
        ax.legend(frameon=False, bbox_to_anchor=(1.02, 1), loc='upper left', title="P (GPa)")
        ax.set_xlim(alpha.index.min(), alpha.index.max())
        ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        fig_name = self.results_folder / 'alpha_T'
        self.save_fig(ax, fig_name)

    def plot_isothermal_bulk_modulus(self):
        f_isothermal_bulk_modulus = self.user_settings['Bt']
        btp = pd.read_csv(f_isothermal_bulk_modulus, sep='\s+', index_col='T(K)\P(GPa)')
        self.plot_to_file()
        ax = btp.plot(y=btp.columns[0], use_index=True, figsize=(6, 5), color='k')
        ax.set_xlabel('Temperature (K)')
        ax.set_ylabel('Isothermal Bulk Modulus (GPa)')
        ax.legend(frameon=True, loc='best', title="P (GPa)")
        ax.set_xlim(btp.index.min(), btp.index.max())
        fig_name = self.results_folder / 'Bt_T'
        self.save_fig(ax, fig_name)

    def plot_pressure_specific_heat_capacity(self):
        f_cp = self.user_settings['Cp']
        cp = pd.read_csv(f_cp, sep='\s+', index_col='T(K)\P(GPa)')
        self.plot_to_file()
        ax = cp.plot(figsize=(6, 5), color=sns.color_palette("Blues", len(cp.columns)))
        ax.set_xlabel('Temperature (K)')
        ax.set_ylabel('Cp (J/mol K)')
        ax.legend(frameon=False, bbox_to_anchor=(1.02, 1), loc='upper left', title="P (GPa)")
        ax.set_xlim(cp.index.min(), cp.index[:-2].max())
        ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        fig_name = self.results_folder / 'Cp_T'
        self.save_fig(ax, fig_name)

    def plot_gruneisen_parameter(self):
        f_gamma = self.user_settings['gamma']
        gamma = pd.read_csv(f_gamma, sep='\s+', index_col='T(K)\P(GPa)')
        self.plot_to_file()
        ax = gamma.plot(figsize=(6, 5), color=sns.color_palette("Greens", len(gamma.columns)))
        ax.set_xlabel('Temperature (K)')
        ax.set_ylabel('Gruneisen Parameter')
        ax.legend(frameon=False, bbox_to_anchor=(1.02, 1), loc='upper left', title="P (GPa)")
        ax.set_xlim(gamma.index.min(), gamma.index[:-2].max())
        ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        fig_name = self.results_folder / 'gamma_T'
        self.save_fig(ax, fig_name)

    def plot_gibbs_free_energy(self):
        f_gibbs = self.user_settings['G']
        gibbs = pd.read_csv(f_gibbs, sep='\s+', index_col='T(K)\P(GPa)')
        self.plot_to_file()
        ax = gibbs.plot(figsize=(6, 5), color=sns.color_palette("binary", len(gibbs.columns)))
        ax.set_xlabel('Temperature (K)')
        ax.set_ylabel('Gibbs Free Energy (ev)')
        ax.legend(frameon=False, bbox_to_anchor=(1.02, 1), loc='upper left', title="P (GPa)")
        ax.set_xlim(gibbs.index.min(), gibbs.index.max())
        fig_name = self.results_folder / 'G_T'
        self.save_fig(ax, fig_name)

    def plot_volume(self):
        f_volume = self.user_settings['V']
        volume = pd.read_csv(f_volume, sep='\s+', index_col='T(K)\P(GPa)')
        self.plot_to_file()
        ax = volume.plot(figsize=(6, 5), color=sns.color_palette("Oranges", len(volume.columns)))
        ax.set_xlabel('Temperature (K)')
        ax.set_ylabel(r'Volume ($\AA^3$)')
        ax.legend(frameon=False, bbox_to_anchor=(1.02, 1), loc='upper left', title="P (GPa)")
        ax.set_xlim(volume.index.min(), volume.index.max())
        fig_name = self.results_folder / 'V_T'
        self.save_fig(ax, fig_name)

    def fv_pv(self):
        temperature_sample: list = self.user_settings['T4FV']
        f_min = []
        v_f_min = []

        f_fitted_volume = self.user_settings['f_tv_fitted']
        volume_tv = pd.read_csv(f_fitted_volume, sep='\s+', index_col='T(K)\V(A^3)')
        volume_tv.index = volume_tv.index.map(str)
        volume_vt = volume_tv.T
        volume_vt.index = volume_vt.index.map(float)
        volume_t = volume_vt[temperature_sample]
        volume_v_grid = np.asarray(volume_vt.index, float)

        f_nonfitted_volume = self.user_settings['f_tv_non_fitted']
        volume_nonfitted_tv = pd.read_csv(f_nonfitted_volume, sep='\s+', index_col='T(K)\V(A^3)')
        volume_nonfitted_tv.index = volume_nonfitted_tv.index.map(str)
        volume_nonfitted_vt = volume_nonfitted_tv.T
        volume_nonfitted_vt.index = volume_nonfitted_vt.index.map(float)
        volume_nonfitted_t = volume_nonfitted_vt[temperature_sample]

        f_p_volume = self.user_settings['p_tv_gpa']
        p_volume_tv = pd.read_csv(f_p_volume, sep='\s+', index_col='T(K)\V(A^3)')
        p_volume_tv.index = p_volume_tv.index.map(str)
        p_volume_vt = p_volume_tv.T
        p_volume_vt.index = p_volume_vt.index.map(float)
        p_volume_t = p_volume_vt[temperature_sample]

        # Get the 'V0' and 'E0' for selected Temperatures
        for t in temperature_sample:
            f_min.append(float(volume_vt[t].min()))
            v_f_min.append(float(volume_vt[t].idxmin()))

        self.plot_to_file()

        fig, axes = plt.subplots(nrows=1, ncols=2)
        volume_t.plot(ax=axes[0], style='b-')
        volume_nonfitted_t.plot(ax=axes[0], style='ko')
        axes[0].plot(v_f_min, f_min, 'ro-', label=None)

        axes[0].set_xlabel(r'Volume ($\AA$)')
        axes[0].set_ylabel('Helmholtz Free Energy (eV)')
        axes[0].legend().set_visible(False)
        axes[0].set_xlim(volume_v_grid.min(), volume_v_grid.max())
        for t in temperature_sample:
            x, y = 1.002 * float(volume_v_grid.max()), volume_vt[t].tolist()[0]
            axes[0].text(x, y, str(t) + ' K', fontsize=8)

        axes[1].axhline(y=self.user_settings['DESIRED_PRESSURES_GPa'][0], color='whitesmoke')
        axes[1].axhline(y=self.user_settings['DESIRED_PRESSURES_GPa'][-1], color='whitesmoke')
        p_volume_t.plot(ax=axes[1], style='k-')

        axes[1].set_xlabel(r'Volume ($\AA$)')
        axes[1].set_ylabel('Pressure (GPa)')
        axes[1].legend().set_visible(False)
        axes[1].set_xlim(volume_v_grid.min(), volume_v_grid.max())

        fig.tight_layout(w_pad=3.2, h_pad=0)
        fig_name_pdf = str(self.results_folder / 'FVT_PVT.pdf')
        fig.savefig(fig_name_pdf, format='pdf', dpi=300, bbox_inches='tight')



#!/usr/bin/env python3

import sys

if sys.version_info < (3, 5):  # In case of user who does not have ``pip`` above version 9.0.0
    raise EnvironmentError('Please use Python version higher than 3.5!')

__author__ = {'Tian Qin': 'qinxx197@umn.edu', 'Qi Zhang': 'qz2280@columbia.edu'}
__copyright__ = 'Copyright (c) 2018, Renata group'
__credits__ = {'Renata M. M. Wentzcovitch': 'rmw2150@columbia.edu'}
__date__ = 'Feb 17, 2018'
__maintainer__ = 'Tian Qin, Qi Zhang'
__version__ = '1.0.14'



#!/usr/bin/env python3
"""
.. module cli.parser
   :platform: Unix, Windows, Mac, Linux
.. moduleauthor:: Chenxing Luo <chenxing.luo@columbia.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

import argparse

import pkg_resources

from qha import __version__
from qha.cli.handler import QHACommandHandler


class SubCommandResolvers:
    """
    This is just a simple dataclass for storing useful parsers that will be dealing with sub-commands.

    :param handler: The handler that will do the corresponding tasks to the sub-command.
    :param parser: The parser that will parse the (optional) arguments of the sub-command.
    """

    def __init__(self, handler, parser):
        self.handler = handler
        self.parser = parser


class QHAArgumentParser:
    def __init__(self):
        self.parser = argparse.ArgumentParser()
        self.init_parser()
        self.subparsers = self.parser.add_subparsers(
            dest='command'
        )
        self.handlers = dict()

    def init_parser(self):
        """
        A function that handles a group of (optional) arguments which will shared with all the sub-commands.
        The available arguments are:

        1. short version: ``-V``, long version: ``--version``, the current release version of ``qha``.
        2. short version: ``-h``, long version: ``--help``, the help document of ``qha``.
        """
        self.parser.add_argument(
            '-V', '--version',
            action='version',
            version="current qha version: {0}".format(__version__),
            help='The current release version of ``qha``.'
        )

    def register_handler(self, command: str, handler: QHACommandHandler, *aliases):
        """
        Each handler corresponds to one sub-command. All sub-commands now available are:

        1. ``convert``
        2. ``run``
        3. ``plot``

        They can be added to the parser and will be parsed.

        :param command: The sub-command's name.
        :param handler: The handler that will do corresponding task, each handler is paired with a sub-command.
        :param aliases: The aliases of the sub-command. It can be an iterable. For example, ``convert`` sub-command
            has an alias that is ``cv``.
        """
        if not isinstance(command, str):
            raise TypeError("Argument *command* should be a string!")

        if not isinstance(handler, QHACommandHandler):
            raise TypeError("Argument *handler* should be a ``QHACommandHandler`` instance!")

        new_parser = self.subparsers.add_parser(command, aliases=list(aliases))
        self.handlers.update(
            dict.fromkeys(
                (command, *aliases),
                SubCommandResolvers(
                    handler=handler,
                    parser=new_parser,
                )
            )
        )
        handler.init_parser(new_parser)

    def parse_args(self, args=None, namespace=None):
        """
        Parse the arguments that is given to ``qha`` command.

        :param args: The (optional) arguments given by the user.
        :param namespace: Exactly the same as the same-name argument in ``argparse.ArgumentParser.parse_args``.
            See the official
            `documentation <https://docs.python.org/3.7/library/argparse.html#argparse.ArgumentParser.parse_args>`_
        :return: Based on argument *namespace*, could be an ``argparse.Namespace`` object or the object specified
            by *namespace*.
        """
        return self.parser.parse_args(args, namespace)

    def invoke_handler(self, namespace):
        """
        Invoke the handler based on *namespace* argument.

        :param namespace: The namespace returned by ``parse_args`` method.
        """
        try:
            command: str = getattr(namespace, 'command')
        except AttributeError:
            raise AttributeError("Argument *namespace* does not have an ``command`` attribute!")

        try:
            handler: QHACommandHandler = self.handlers[command].handler
            handler.run(namespace)
        except KeyError:
            raise ValueError("Command '{0}' is not recognized!".format(command))

    def load_plugins(self):
        """
        Load plugins. Leave for the future plugin system.
        """
        for plugin in pkg_resources.iter_entry_points(group='qha.plugins'):
            klass = plugin.load()
            aliases = klass.aliases if 'aliases' in dir(klass) else None
            self.register_handler(plugin.name, klass(), *aliases)



#!/usr/bin/env python3
"""
.. module unit_conversion
   :platform: Unix, Windows, Mac, Linux
   :synopsis: A module of packaged unit-conversion functions, with [Numba's](https://numba.pydata.org) ``@jit`` speedup.
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

from numba import float64, vectorize
from scipy.constants import physical_constants, electron_volt, angstrom, Avogadro

from qha.settings import DEFAULT_SETTINGS

# ===================== What can be exported? =====================
__all__ = [
    'j_to_ev',
    'ev_to_j',
    'gpa_to_megabar',
    'megabar_to_gpa',
    'b3_to_a3',
    'a3_to_b3',
    'eh_to_ev',
    'ev_to_eh',
    'ry_to_ev',
    'ev_to_ry',
    'j_to_eh',
    'eh_to_j',
    'eh_to_hz',
    'hz_to_eh',
    'eh_to_k',
    'k_to_eh',
    'eh_to_m_inverse',
    'm_inverse_to_eh',
    'eh_to_cm_inverse',
    'cm_inverse_to_eh',
    'ev_to_m_inverse',
    'm_inverse_to_ev',
    'ev_to_cm_inverse',
    'cm_inverse_to_ev',
    'ev_to_k',
    'k_to_ev',
    'ry_to_j',
    'j_to_ry',
    'gpa_to_ev_a3',
    'ev_a3_to_gpa',
    'gpa_to_ry_b3',
    'ry_b3_to_gpa',
    'gpa_to_ha_b3',
    'ha_b3_to_gpa',
    'ev_b3_to_gpa',
    'gpa_to_ev_b3',
    'ry_b_to_ev_a',
    'ha_b_to_ev_a',
    'ry_to_kj_mol',
    'ry_to_j_mol'
]

# ===================== Constants =====================
BOHR = physical_constants['Bohr radius'][0]
EH_EV = physical_constants['Hartree energy in eV'][0]
RY_EV = physical_constants['Rydberg constant times hc in eV'][0]
EH_J = physical_constants['hartree-joule relationship'][0]
EH_HZ = physical_constants['hartree-hertz relationship'][0]
EH_K = physical_constants['hartree-kelvin relationship'][0]
EH_M_INVERSE = physical_constants['hartree-inverse meter relationship'][0]
EV_M_INVERSE = physical_constants['electron volt-inverse meter relationship'][0]
EV_K = physical_constants['electron volt-kelvin relationship'][0]
RY_J = physical_constants['Rydberg constant times hc in J'][0]

# ===================== Settings =====================
_target = DEFAULT_SETTINGS['target']


# ===================== Functions =====================
@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def j_to_ev(value):
    """
    Convert the *value* in unit joule to electronvolt.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / electron_volt


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ev_to_j(value):
    """
    Convert the *value* in unit electronvolt to joule.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * electron_volt


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def gpa_to_megabar(value):
    """
    Convert the *value* in unit gigapascal to megabar.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * 0.01


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def megabar_to_gpa(value):
    """
    Convert the *value* in unit megabar to gigapascal.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / 0.01


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def b3_to_a3(value):
    """
    Convert the *value* in unit cubic bohr radius to what in cubic angstrom.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * (BOHR / angstrom) ** 3


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def a3_to_b3(value):
    """
    Convert the *value* in unit cubic angstrom to what in cubic bohr radius.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * (angstrom / BOHR) ** 3


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def eh_to_ev(value):
    """
    Convert the *value* in unit hartree to electronvolt.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * EH_EV


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ev_to_eh(value):
    """
    Convert the *value* in unit electronvolt to hartree.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / EH_EV


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ry_to_ev(value):
    """
    Convert the *value* in unit rydberg to electronvolt.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * RY_EV


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ev_to_ry(value):
    """
    Convert the *value* in unit electronvolt to rydberg.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / RY_EV


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def j_to_eh(value):
    """
    Convert the *value* in unit joule to hartree.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / EH_J


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def eh_to_j(value):
    """
    Convert the *value* in unit hartree to joule.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * EH_J


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def eh_to_hz(value):
    """
    Convert the *value* in unit hartree to hertz.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * EH_HZ


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def hz_to_eh(value):
    """
    Convert the *value* in unit hertz to hartree.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / EH_HZ


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def eh_to_k(value):
    """
    Convert the *value* in unit hartree to kelvin.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * EH_K


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def k_to_eh(value):
    """
    Convert the *value* in unit kelvin to hartree.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / EH_K


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def eh_to_m_inverse(value):
    """
    Convert the *value* in unit hartree to :math:`\\text{m}^{-1}`.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * EH_M_INVERSE


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def m_inverse_to_eh(value):
    """
    Convert the *value* in unit :math:`\\text{m}^{-1}` to hartree.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / EH_M_INVERSE


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def eh_to_cm_inverse(value):
    """
    Convert the *value* in unit hartree to :math:`\\text{cm}^{-1}`.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * EH_M_INVERSE / 100


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def cm_inverse_to_eh(value):
    """
    Convert the *value* in unit :math:`\\text{cm}^{-1}` to hartree.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / EH_M_INVERSE * 100


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ev_to_m_inverse(value):
    """
    Convert the *value* in unit electronvolt to :math:`\\text{m}^{-1}`.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * EV_M_INVERSE


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def m_inverse_to_ev(value):
    """
    Convert the *value* in unit :math:`\\text{m}^{-1}` to electronvolt.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / EV_M_INVERSE


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ev_to_cm_inverse(value):
    """
    Convert the *value* in unit electronvolt to :math:`\\text{cm}^{-1}`.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * EV_M_INVERSE / 100


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def cm_inverse_to_ev(value):
    """
    Convert the *value* in unit :math:`\\text{cm}^{-1}` to electronvolt.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / EV_M_INVERSE * 100


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ev_to_k(value):
    """
    Convert the *value* in unit electronvolt to kelvin.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * EV_K


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def k_to_ev(value):
    """
    Convert the *value* in unit kelvin to electronvolt.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / EV_K


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ry_to_j(value):
    """
    Convert the *value* in unit rydberg to joule.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * RY_J


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def j_to_ry(value):
    """
    Convert the *value* in unit joule to rydberg.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / RY_J


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def gpa_to_ev_a3(value):
    """
    Convert the *value* in unit gigapascal to :math:`\\frac{ \\text{electronvolt} }{ \\text{angstrom}^3 }`.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * 1e9 / electron_volt * angstrom ** 3


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ev_a3_to_gpa(value):
    """
    Convert the *value* in unit :math:`\\frac{ \\text{electronvolt} }{ \\text{angstrom}^3 }` to gigapascal.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / 1e9 * electron_volt / angstrom ** 3


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def gpa_to_ev_b3(value):
    """
    Convert the *value* in unit gigapascal to :math:`\\frac{ \\text{electronvolt} }{ \\text{bohr radius}^3 }`.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * 1e9 / electron_volt * BOHR ** 3


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ev_b3_to_gpa(value):
    """
    Convert the *value* in unit :math:`\\frac{ \\text{electronvolt} }{ \\text{bohr radius}^3 }` to gigapascal.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / 1e9 * electron_volt / BOHR ** 3


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def gpa_to_ry_b3(value):
    """
    Convert the *value* in unit gigapascal to :math:`\\frac{ \\text{rydberg} }{ \\text{bohr radius}^3 }`.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * 1e9 / RY_J * BOHR ** 3


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ry_b3_to_gpa(value):
    """
    Convert the *value* in unit :math:`\\frac{ \\text{rydberg} }{ \\text{bohr radius}^3 }` to gigapascal.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value / 1e9 * RY_J / BOHR ** 3


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def gpa_to_ha_b3(value):
    """
    Convert the *value* in unit gigapascal to :math:`\\frac{ \\text{hartree} }{ \\text{bohr radius}^3 }`.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * 1e9 / RY_J * BOHR ** 3 / 2


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ha_b3_to_gpa(value):
    """
    Convert the *value* in unit :math:`\\frac{ \\text{hartree} }{ \\text{bohr radius}^3 }` to gigapascal.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return 2 * value / 1e9 * RY_J / BOHR ** 3


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ry_b_to_ev_a(value):
    """
    The atomic force.
    Convert the *value* in unit :math:`\\frac{ \\text{rydberg} }{ \\text{bohr radius} }` to
    :math:`\\frac{ \\text{electronvolt} }{ \\text{angstrom} }`.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * RY_EV / (BOHR / angstrom)


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ha_b_to_ev_a(value):
    """
    The atomic force.
    Convert the *value* in unit :math:`\\frac{ \\text{hartree} }{ \\text{bohr radius} }` to
    :math:`\\frac{ \\text{electronvolt} }{ \\text{angstrom} }`.

    :param value: The value to be converted.
    :return: The converted value.
    """
    return value * EH_EV / (BOHR / angstrom)


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ry_to_kj_mol(value):
    """
    Convert the *value* is in unit Rydberg, the converted value is in unit
    :math:`\\frac{ \\text{k J} }{ \\text{mol} }`.

    :param value: The value to be converted.
    :return: the converted value.
    """
    return value * Avogadro * RY_J / 1000


@vectorize([float64(float64)], nopython=True, cache=True, target=_target)
def ry_to_j_mol(value):
    """
    Convert the *value* is in unit Rydberg, the converted value is in unit
    :math:`\\frac{ \\text{k J} }{ \\text{mol} }`.

    :param value: The value to be converted.
    :return: the converted value.
    """
    return value * Avogadro * RY_J



#!/usr/bin/env python3

import codecs
import os
import re
from distutils.core import setup

import setuptools

# Referenced from `here <https://packaging.python.org/guides/single-sourcing-package-version/>`_.
here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(name='qha',
      version=find_version('qha', '__init__.py'),
      description='A powerful tool for quasi-harmonic approximation',
      author='Tian Qin, Qi Zhang',
      author_email='qinxx197@umn.edu, qz2280@columbia.edu',
      maintainer='Tian Qin, Qi Zhang',
      maintainer_email='qinxx197@umn.edu, qz2280@columbia.edu',
      license='GNU General Public License 3',
      url='https://github.com/MineralsCloud/qha',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering :: Physics',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python :: 3.6',
      ],
      python_requires='>=3.6, <3.7',
      keywords='thermodynamic-properties quasi-harmonic-approximation scientific-computation',
      install_requires=[
          'lazy_property',
          'numba',
          'numpy',
          'bigfloat',
          'pandas',
          'scientific-string',
          'scipy',
          'text-stream',
          'pyyaml',
          'matplotlib',
          'seaborn',
      ],
      packages=[
          'qha',
          'qha.multi_configurations',
          'qha.cli',
          'qha.basic_io',
      ],
      entry_points={
          'console_scripts': [
              'qha=qha.cli:main'
          ],
      })



#!/usr/bin/env python3
"""
.. module grid_interpolation
   :platform: Unix, Windows, Mac, Linux
   :synopsis: This module defines a class ``FinerGrid`` that will do Birch--Murnaghan EoS fitting on the input
    free energies and volumes, and evaluate the fitted function on a denser volume grid.
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

from typing import Optional, Tuple

import numpy as np
from numba import vectorize, float64

from qha.fitting import polynomial_least_square_fitting, apply_finite_strain_fitting
from qha.type_aliases import Vector, Matrix
from qha.unit_conversion import gpa_to_ry_b3

# ===================== What can be exported? =====================
__all__ = [
    'calculate_eulerian_strain',
    'from_eulerian_strain',
    'VolumeExpander',
    'FinerGrid'
]


@vectorize([float64(float64, float64)], nopython=True)
def calculate_eulerian_strain(v0, vs):
    """
    Calculate the Eulerian strain (:math:`f`s) of a given volume vector *vs* with respect to a reference volume *v0*,
    where

    .. math::

       f = \\frac{ 1 }{ 2 } \\bigg( \\Big( \\frac{ V_0 }{ V }^{2/3} \\Big) -1 \\bigg).

    :param v0: The volume set as the reference for the Eulerian strain calculation.
    :param vs: A volume vector, each item of which will be calculated the Eulerian strain with respect to *v0*.
    :return: A vector of the calculated Eulerian strain.
    """
    return 1 / 2 * ((v0 / vs) ** (2 / 3) - 1)


@vectorize([float64(float64, float64)], nopython=True)
def from_eulerian_strain(v0, fs):
    """
    Calculate the corresponding volumes :math:`V`s from a vector of given Eulerian strains (*fs*)
    and a reference volume *v0*. It is the inverse function of ``calculate_eulerian_strain``, i.e.,

    .. math::

       V = V_0 (2 f + 1)^{-3/2}.

    :param v0: The volume set as a reference for volume calculation, i.e., :math:`V_0` mentioned above.
    :param fs: A vector of Eulerian strains with respect to :math:`V_0`.
    :return: A vector of calculated volume :math:`V`s.
    """
    return v0 * (2 * fs + 1) ** (-3 / 2)


class VolumeExpander:
    """
    Interpolate volumes on input volumes *in_volumes*, with *ratio* given.
    For the Eulerian strain, the larger the strain, the smaller the volume.
    So larger volume corresponds to smaller strain.

    Algorithm:

    1. Find the maximum and minimum volumes of *in_volumes*.
    2. Expand the lower and upper bounds of the volumes to :math:`V_\\text{lower}` and :math:`V_\\text{upper}` by :math:`V_\\text{lower} = \\frac{ V_\\text{min} }{ r }` and :math:`V_\\text{upper} = V_\\text{max} r`, where :math:`r` is the *ratio*, which is usually an empirical parameter.
    3. The Eulerian strains of :math:`V_\\text{lower}` and :math:`V_\\text{upper}` with respect to :math:`V_\\text{max}` is calculated. Then a linear sequence between the lower and upper bounds of the strains is generated. It is the ``strains`` attribute.
    4. The volumes corresponding to ``strains`` are generated by function ``from_eulerian_strain``.

    **Note:** To use this class, the method ``interpolate_volumes`` must be called immediately after
    instantiating it. Otherwise, the ``strains`` and ``out_volumes`` attributes are all ``None``, which is usually not
    expected.

    :param in_volumes: An input vector of volumes.
    :param out_volumes_num: Number of output volumes. It should be larger than the number of input volumes.
    :param ratio: The ratio of the upper bounds of expanded volumes with respect to the largest input volume in
        *in_volumes*. An empirical parameter, better lower than ``1.5``. Otherwise, the later
        result may not be reliable.
    """

    def __init__(self, in_volumes: Vector, out_volumes_num: int, ratio: float):
        self._in_volumes = np.array(in_volumes, dtype=float)
        self._out_volumes_num = int(out_volumes_num)
        self._ratio = float(ratio)
        self._strains = None
        self._out_volumes = None

    @property
    def in_volumes(self):
        """
        :return: Input volume vector, read only.
        """
        return self._in_volumes

    @property
    def ratio(self) -> float:
        """
        An empirical parameter, usually larger than ``1`` and lower than ``1.5``.

        :return: A floating-point number described in the algorithm above.
        """
        return self._ratio

    @ratio.setter
    def ratio(self, value: float):
        self._ratio = float(value)

    @property
    def out_volumes_num(self) -> int:
        """
        :return: The number of output volumes user wants, usually greater than the number of input volumes.
        """
        return self._out_volumes_num

    @out_volumes_num.setter
    def out_volumes_num(self, value: int):
        if not isinstance(value, int):
            raise TypeError("The argument *out_volumes_num* must be an integer!")
        if value <= 0:
            raise ValueError("The argument *out_volumes_num* must be an integer larger than 0!")
        self._out_volumes_num = value

    @property
    def strains(self) -> Optional[Vector]:
        """
        An arithmetic sequence starting from the lower bound to the upper bound of the expanded strain,
        the number of elements in it is specified by ``out_volumes_num`` attribute.

        :return: A vector of strains if method ``interpolate_volumes`` is called, otherwise ``None``
            is returned.
        """
        return self._strains

    @property
    def out_volumes(self) -> Optional[Vector]:
        """
        The volumes interpolated by the algorithm described above.

        :return: If method ``interpolate_volumes`` is called, a vector of output volumes is returned, otherwise ``None``
            is returned.
        """
        return self._out_volumes

    def interpolate_volumes(self) -> None:
        """
        The algorithm is described above. This method should be called immediately after an instance of the
        class is created, unless user is clear what is being done.
        """
        v_min, v_max = np.min(self._in_volumes), np.max(self._in_volumes)
        # r = v_upper / v_max = v_min / v_lower
        v_lower, v_upper = v_min / self._ratio, v_max * self._ratio
        # The *v_max* is a reference value here.
        s_upper, s_lower = calculate_eulerian_strain(v_max, v_lower), calculate_eulerian_strain(v_max, v_upper)
        self._strains = np.linspace(s_lower, s_upper, self._out_volumes_num)
        self._out_volumes = from_eulerian_strain(v_max, self._strains)


class FinerGrid:
    """
    A class that will do the Birch--Murnaghan finite-strain equation of state fitting,
    moreover, evaluate the free energies on the denser volume vector.

    :param desired_p_min: The desired minimum pressure to calculate for further steps.
    :param dense_volumes_amount: The number of volumes on a denser grid.
    :param order: The order of the Birch--Murnaghan finite-strain equation of state fitting.
    """

    def __init__(self, desired_p_min: float, dense_volumes_amount: int, order: Optional[int] = 3):
        self.desired_p_min = float(desired_p_min)
        self.dense_volumes_amount = int(dense_volumes_amount)
        self.option = int(order)
        self._ratio = None

    @property
    def ratio(self) -> float:
        return self._ratio

    def approach_to_best_ratio(self, volumes: Vector, free_energies: Vector, initial_ratio: float) -> float:
        """
        Trying to find the best volume grids based on an a very large volume grids.

        :param volumes: Volumes of these calculations were perform (sparse).
        :param free_energies: Free energies at the highest temperature (sparse).
        :param initial_ratio: Initial ratio, a guess value, which can be set to a very large number.
        :return: The suitable `ratio` for further calculation.
        """
        vr = VolumeExpander(volumes, self.dense_volumes_amount, initial_ratio)
        vr.interpolate_volumes()
        strains, finer_volumes = vr.strains, vr.out_volumes
        eulerian_strain = calculate_eulerian_strain(volumes[0], volumes)
        _, f_v_tmax = polynomial_least_square_fitting(eulerian_strain, free_energies, strains, self.option)
        p_v_tmax = -np.gradient(f_v_tmax) / np.gradient(finer_volumes)
        p_desire = gpa_to_ry_b3(self.desired_p_min)
        # Find the index of the first pressure value that slightly smaller than p_desire.
        idx = np.argmin(p_v_tmax < p_desire)
        final_ratio = finer_volumes[idx] / max(volumes)
        return final_ratio

    def refine_grid(self, volumes: Vector, free_energies: Matrix,
                    ratio: Optional[float] = None) -> Tuple[Vector, Matrix, float]:
        """
        Get the appropriate volume grid for interpolation.
        Avoid to use a too large volume grid to obtain data, which might lose accuracy.

        :param free_energies: Calculated Helmholtz Free Energies for input volumes (sparse).
        :param volumes: Volumes of these calculations were performed (sparse).
        :param ratio: This ratio is used to get a larger volume grid
        :return: volume, Helmholtz free energy at a denser vector, and the `ratio` used in this calculation
        """
        if ratio is not None:
            new_ratio: float = ratio
        else:
            new_ratio = self.approach_to_best_ratio(volumes, free_energies[-1, :], 1.45)
            if new_ratio < 1.0:
                # If the ``new_ratio`` is smaller than 1.0, which means the volumes calculated is large enough,
                # there is no need to expand the volumes.
                new_ratio = 1.0

        self._ratio = new_ratio

        eulerian_strain = calculate_eulerian_strain(volumes[0], volumes)
        vr = VolumeExpander(in_volumes=volumes, out_volumes_num=self.dense_volumes_amount, ratio=new_ratio)
        vr.interpolate_volumes()  # As mentioned in ``VolumeExpander`` doc, call this method immediately.
        strains, dense_volumes = vr.strains, vr.out_volumes
        dense_free_energies = apply_finite_strain_fitting(eulerian_strain, free_energies, strains, self.option)
        return dense_volumes, dense_free_energies, new_ratio



#!/usr/bin/env python3
"""
.. module statmech
   :platform: Unix, Windows, Mac, Linux
   :synopsis:
.. moduleauthor:: Qi Zhang <qz2280@columbia.edu>
"""

import numpy as np
from numba import float64, vectorize
from scipy.constants import physical_constants as pc

import qha.settings

# ===================== What can be exported? =====================
__all__ = ['ho_free_energy', 'subsystem_partition_function', 'log_subsystem_partition_function']

K = {'ha': pc['Boltzmann constant in eV/K'][0] / pc['Hartree energy in eV'][0],
     'ry': pc['Boltzmann constant in eV/K'][0] / pc['Rydberg constant times hc in eV'][0],
     'ev': pc['Boltzmann constant in eV/K'][0],
     'SI': pc['Boltzmann constant'][0]}[qha.settings.energy_unit]

HBAR = {'ha': 100 / pc['electron volt-inverse meter relationship'][0] / pc['Hartree energy in eV'][0],
        'ry': 100 / pc['electron volt-inverse meter relationship'][0] / pc['Rydberg constant times hc in eV'][0],
        'ev': 100 / pc['electron volt-inverse meter relationship'][0],
        'SI': 100 / pc['electron volt-inverse meter relationship'][0] / pc['joule-electron volt relationship'][0]}[
    qha.settings.energy_unit]


@vectorize([float64(float64, float64)], nopython=True, target='parallel', cache=True)
def ho_free_energy(temperature, frequency):
    """
    Calculate Helmholtz free energy of a single harmonic oscillator at a specific temperature.
    This is a vectorized function so the argument *frequency* can be an array.

    :param temperature: The temperature, in unit 'Kelvin'. Zero-temperature is allowed.
    :param frequency: The frequency of the harmonic oscillator, in unit 'per cm'. If the *frequency*
        is less than or equal to :math:`0`, directly return ``0`` as its free energy.
    :return: Helmholtz free energy of the harmonic oscillator, whose unit depends on user's settings.
    """
    if frequency <= 0:  # Contribution from Gamma point's acoustic frequencies is zero.
        return 0

    hw = HBAR * frequency
    kt = K * temperature
    return 1 / 2 * hw + kt * np.log(1 - np.exp(-hw / kt))


@vectorize([float64(float64, float64)], nopython=True, target='parallel', cache=True)
def subsystem_partition_function(temperature, frequency):
    """
    Calculate the subsystem partition function of a single harmonic oscillator at a specific temperature.
    This is a vectorized function so the argument *frequency* can be an array.

    :param temperature: The temperature, in unit 'Kelvin'. Zero-temperature is allowed.
    :param frequency: The frequency of the harmonic oscillator, in unit 'per cm'. If the *frequency*
        is less than or equal to :math:`0`, directly return ``1`` as its subsystem partition function value.
    :return: The subsystem partition function of the harmonic oscillator.
    """
    if frequency <= 0:
        return 1

    x = -HBAR * frequency / (K * temperature)
    return np.exp(x / 2) / (1 - np.exp(x))


@vectorize([float64(float64, float64)], nopython=True, target='parallel', cache=True)
def log_subsystem_partition_function(temperature, frequency):
    """
    Calculate the natural logarithm of the subsystem partition function of a single harmonic oscillator
    at a specific temperature. This is a vectorized function so the argument *frequency* can be an array.

    :param temperature: The temperature, in unit 'Kelvin'. Zero-temperature is allowed.
    :param frequency: The frequency of the harmonic oscillator, in unit 'per cm'. If the *frequency*
        is less than or equal to :math:`0`, directly return ``0`` as a result.
    :return: Natural logarithm of the subsystem partition function of the oscillator.
    """
    if frequency <= 0:
        return 0

    x = -HBAR * frequency / (K * temperature)
    return x / 2 - np.log(1 - np.exp(x))



#!/usr/bin/env python3
"""
.. module fitting
   :platform: Unix, Windows, Mac, Linux
   :synopsis: This module is one of the most important part of this package since it implements a robust
    finite strain EoS fitting for ``grid_interpolation`` module's use.
.. moduleauthor:: Tian Qin <qinxx197@umn.edu>
"""

from typing import Optional

import numpy as np
from numpy.linalg import inv
from qha.type_aliases import Matrix, Vector

# ===================== What can be exported? =====================
__all__ = ['polynomial_least_square_fitting', 'apply_finite_strain_fitting']


def polynomial_least_square_fitting(xs, ys, new_xs, order: Optional[int] = 3):
    """
    The algorithm is referenced from the
    `Wolfram MathWorld <http://mathworld.wolfram.com/LeastSquaresFittingPolynomial.html>`_.

    :param xs: A vector of existing x-coordinates.
    :param ys: A vector of y-coordinates correspond to the *xs*.
    :param new_xs: A new vector of x-coordinates to be applied with the polynomial-fitting result.
    :param order: The order chose to fit the finite strain EoS, the default value is ``3``,
        which is, the third-order Birch--Murnaghan EoS.
    :return: A tuple, the polynomial-fitting coefficients and the new vector of y-coordinates.
    """
    order += 1  # The definition of order in ``numpy.vander`` is different from the order in finite strain by one.
    xx = np.vander(xs, order, increasing=True)  # This will make a Vandermonde matrix that will be used in EoS fitting.
    xx_t = xx.T  # Transpose the matrix.
    a = inv(xx_t @ xx) @ xx_t @ ys  # a = (X^T . X)^{-1} . X^T . ys
    new_y = np.vander(new_xs, order, increasing=True) @ a
    return a, new_y


def apply_finite_strain_fitting(strains_sparse: Vector, free_energies: Matrix, strains_dense: Vector,
                                order: Optional[int] = 3):
    """
    Calculate the free energies :math:`F(T, V)` for some strains (*strains_dense*), with the
    free energies (*free_energies*) on some other strains (*strains_sparse*) known already.
    Do a polynomial curve-fitting the apply the fitted function
    to the *strains_dense*.

    :param strains_sparse: A vector of the Eulerian strains for a sparse set of volumes.
    :param free_energies: The free energies correspond to *strains_sparse* at several temperature.
    :param strains_dense: A vector of the Eulerian strains at a denser set of volumes.
    :param order: The order chose to fit the finite strain EoS, the default value is ``3``,
        which is, the third-order Birch--Murnaghan EoS.
    :return: The free energies correspond to *strains_dense* at different temperature.
    """
    temperature_amount, _ = free_energies.shape
    dense_volume_amount = len(strains_dense)
    f_v_t = np.empty((temperature_amount, dense_volume_amount))  # Initialize the F(T,V) array

    for i in range(temperature_amount):
        _, f_i = polynomial_least_square_fitting(strains_sparse, free_energies[i], strains_dense, order)
        f_v_t[i] = f_i
    return f_v_t
